package sdf;

import java.text.DecimalFormatSymbols;
import java.util.Arrays;
import java.util.Comparator;
import java.util.Enumeration;
import java.util.Hashtable;

import nenok.va.Domain;
import nenok.va.Idempotency;
import nenok.va.Identity;
import nenok.va.Predictor;
import nenok.va.Regularity;
import nenok.va.VAException;
import nenok.va.Valuation;
import nenok.va.Variable;

/**
 * Shortest distance functions are an example of a valuation algebra and this
 * class is a possible implementation of the valuation interface for the case
 * of shortest distance functions. Internally it is represented by a two-
 * dimensional hashtable.
 * 
 * @author Antoine De Groote, Marc Pouly
 * @version 1.1
 */

public class SDF implements Valuation, Idempotency {
	
    //Distance for equality checks:
    private static final double EPSILON = 0.0000001;
	
	private SDF_Variable[] variables;
	private Hashtable<SDF_Variable, Hashtable<SDF_Variable, Double>> values;
	private Domain domain;
	
	/*
	 * Special symbol table:
	 */
	
	private static final DecimalFormatSymbols SYMBOLS = new DecimalFormatSymbols();
	
	/*
	 * Constructors:
	 */
	
	/**
	 * Constructor:
	 * @param variables The nodes of the network that this distance function covers. 
	 * @param values The values for each possible link. (bidimensional array)
	 * If variables is of size n, then values must be of size n*n.
	 */
	
	public SDF(SDF_Variable[] variables, double[][] values) {
		if(variables == null || values == null) {
			throw new IllegalArgumentException("Assert arguments being != null.");
		}
		setInstanceFields(variables, doubleToHash(variables, values));
	}
	
	/**
	 * Constructor:
	 * @param variables The nodes of the network that this distance function covers. 
	 * @param values The values for each possible link. (bidimensional hashtable)
	 * If variables is of size n, then values must be of size n*n.
	 */
	
	protected SDF(SDF_Variable[] variables, Hashtable<SDF_Variable, Hashtable<SDF_Variable, Double>> values) {
		if(variables == null || values == null) {
			throw new IllegalArgumentException("Assert arguments being != null.");
		}
		setInstanceFields(variables, values);
	}
			
	/**
	 * Sets instance fields (variables, values and domain) 
	 * and performs dimension checks.
	 * @param variables An array of discrete variables.
	 * @param values A bidimensional hashtable of distances.
	 */

	@SuppressWarnings("cast") 
	private void setInstanceFields(SDF_Variable[] variables, Hashtable<SDF_Variable, Hashtable<SDF_Variable, Double>> values) {
		this.variables = variables;
		this.values = values;
		this.domain = new Domain((Variable[]) variables);
		if (domain.size() != variables.length) {
			throw new IllegalArgumentException("Multiply defined variables. Make sure all variables are distinct.");
		}
		if (values.size() != variables.length) {
			throw new IllegalArgumentException("Dimension of values table doesn't match. Given: "+values.size()+", expected: "+variables.length+".");
		}
		for (Variable var : domain) {
			if (! values.containsKey(var)) {
				throw new IllegalArgumentException("Domain variable "+var.toString()+" not contained in values table.");
			}
			if (values.get(var).size() != variables.length) {
				throw new IllegalArgumentException("Dimension of internal table row of variable "+var.toString()+" of values table doesn't match. Given: "+values.get(var).size()+", expected: "+variables.length+".");
			}
			if (Math.abs(values.get(var).get(var)) > 0.00001) {
				throw new IllegalArgumentException("Some diagonal element is not equal to 0: "+var.toString()+","+var.toString()+" : "+values.get(var).get(var));
			}
		}
	}
	
	/**
	 * Converts a bidimensional array into a bidimensional hashtable.
	 * @param variables The variables for the associated SDF.
	 * @param values The bidimensional array containing the distances.
	 * @return A bidimensional hashtable containing the distances.
	 */
	
	private Hashtable<SDF_Variable, Hashtable<SDF_Variable, Double>> doubleToHash(SDF_Variable[] variables, double[][] values) {
		Hashtable<SDF_Variable, Hashtable<SDF_Variable, Double>> result = new Hashtable<SDF_Variable, Hashtable<SDF_Variable, Double>>();
		for (int i = 0; i < variables.length; i++) {
			Hashtable<SDF_Variable, Double> cur = new Hashtable<SDF_Variable, Double>();
			for (int j = 0; j < variables.length; j++) {
				cur.put(variables[j], values[i][j]);
			}
			result.put(variables[i], cur);
		}
		return result;
	}
	
	/*
	 * Valuation implementation:
	 */
	
	/**
	 * Returns the domain of this SDF.
	 * @see nenok.va.Valuation#label()
	 */
	
	public Domain label() {
		return domain;
	}
	
	/**
	 * Combines this SDF with val and returns a new SDF object.
	 * @param val the SDF this SDF is to be combined with
	 * @return the resulting, newly created SDF
	 * @see nenok.va.Valuation#combine(nenok.va.Valuation)
	 */
	
	public Valuation combine(final Valuation val) {
		if(val instanceof Identity) {
			return this;
		}
		SDF arg = (SDF)val;
		
		Domain union = Domain.union(arg.domain, this.domain);
		
		/*
		 * Calculate the union of the two distance functions
		 */
		
		SDF d1 = (SDF) this.extension(union);
		SDF d2 = (SDF) arg.extension(union);
		
		for (Variable var1 : union) {
			for (Variable var2 : union) {
				SDF_Variable dfvar2 = (SDF_Variable) var2;
				d1.values.get(var1).put(dfvar2, Math.min(d1.values.get(var1).get(var2),
						d2.values.get(var1).get(var2)));
			}
		}
		
		/*
		 * Minimalize the union obtained above using the iterative exponential algorithm.
		 */
		
		double min = Double.POSITIVE_INFINITY;
		Hashtable<SDF_Variable, Hashtable<SDF_Variable, Double>> ah = deepClone(d1.values);
		Hashtable<SDF_Variable, Hashtable<SDF_Variable, Double>> ah2;
		for (int m = 0; m < Math.ceil(Math.log(d1.domain.size())/Math.log(2)); m++) {
			ah2 = deepClone(ah);
			for (Variable i : union) {
				for (Variable j : union) {
					min = Double.POSITIVE_INFINITY;
					for (Variable k : union) {
						double cur = ah2.get(i).get(k) + ah2.get(k).get(j);
						if (cur < min) {
							min = cur;
						}
					}
					ah.get(i).put((SDF_Variable)j, min);
				}
			}
		}
		
		return new SDF(d1.variables, ah);
	}
	
	/**
	 * Deep clones a bidimensional hashtable.
	 * @param arg The array to be deep cloned.
	 * @return A deep copy of a bidimensional hashtable.
	 */
	
	private Hashtable<SDF_Variable, Hashtable<SDF_Variable, Double>> deepClone(Hashtable<SDF_Variable, Hashtable<SDF_Variable, Double>> arg) {
		Hashtable<SDF_Variable, Hashtable<SDF_Variable, Double>> result = new Hashtable<SDF_Variable, Hashtable<SDF_Variable, Double>>();
		for (Enumeration<SDF_Variable> e = arg.keys(); e.hasMoreElements();) {
			SDF_Variable var = e.nextElement(); //level 1
			Hashtable<SDF_Variable, Double> newHashtable = new Hashtable<SDF_Variable, Double>();
			for (Enumeration<SDF_Variable> ee = arg.get(var).keys(); ee.hasMoreElements();) {
				SDF_Variable var2 = ee.nextElement(); //level 2
				newHashtable.put(var2, arg.get(var).get(var2));
			}
			result.put(var, newHashtable); 
		}
		return result;
	}
	
	/**
	 * Marginalizes this SDF to the domain dom. Row and columns (i.e. hashtable
	 * entries) that don't appear in dom are dropped, and a new SDF object is returned.
	 * @param dom the domain this SDF is to be projected to
	 * @return a new, projected SDF object
	 * @see nenok.va.Valuation#marginalize(nenok.va.Domain)
	 */
	
	public Valuation marginalize(final Domain dom) {
		if(domain.equals(dom)) {
			return this;
		}
		SDF result = this;
		if(!dom.subSetOf(domain)) {
			new VAException("Marginalization impossible: this.label() = "+domain+", argument = "+dom+".");
		} 
		//Calculates the difference set between both domains.
		Domain difference = Domain.difference(domain, dom);
		//Remove each variable in the difference set.
		for (Variable var : difference) {
			result = result.eliminateVariable((SDF_Variable)var);
		}
		return result;
	}
	
	/**
	 * Calculates the vacuous extension of this SDF to the specified domain.
	 * @param dom The domain to which this SDF is to be extended.
	 * @return an SDF with domain dom
	 */
	
	public Valuation extension(final Domain dom) {
		if(domain.equals(dom)) {
			return this;
		}
		SDF result = (SDF)this.clone();
		if(!domain.subSetOf(dom)) {
			new VAException("Vacuous extension impossible: this.label() = "+dom+", argument = "+domain+".");
		} 
		//Calculates the difference set between both domains.
		Domain difference = Domain.difference(dom, domain);
		
		//Create the new Variables
		SDF_Variable[] newVariables = new SDF_Variable[dom.size()];
		int i = 0;
		for (Variable var : dom) {
			newVariables[i++] = (SDF_Variable) var;
		}
		
		//Insert missing rows 
		for (Variable var : difference) {
			result.values.put((SDF_Variable) var, new Hashtable<SDF_Variable, Double>());
		}
		
		//Create the new values, i.e. fill in 0's and inifinities (Double.POSITIVE_INFINITY).
		for (Variable var : difference) {
			for (Variable var2 : newVariables) {
				result.values.get(var).put((SDF_Variable) var2, Double.POSITIVE_INFINITY);
				result.values.get(var2).put((SDF_Variable) var, Double.POSITIVE_INFINITY);
			}
			result.values.get(var).put((SDF_Variable) var, 0.0);
		}
		
		return new SDF(newVariables, result.values);
	}
	
	/**
	 * Clones this SDF.
	 * @return a copy of this SDF
	 * @see java.lang.Object#clone()
	 */
	
	public Object clone() {
		
		/*
		 * There is no need to clone the variable array. 
		 * Variables can be shared among distance functions.
		 */
		
		return new SDF(this.variables, deepClone(this.values));
	}
	
	/**
	 * @see nenok.va.Valuation#weight()
	 */
	
	public int weight() {
		return predictor().predict(this.label());
	}
	
	/*
	 * Regularity implementation:
	 */
	
	/**
     * @see nenok.va.Regularity#inverse()
     */
    
    public Regularity inverse() {
        return new Idempotency.Implementor().inverse(this);
    }
	
	/*
	 * Predictability implementation:
	 */
	
	/**
	 * @return The prdictor of this valuation algebra instance.
	 * @see nenok.va.Predictability#predictor()
	 */ 
	
	public Predictor predictor() {
		return SDF_Predictor.getInstance();
	}
	
	/*
	 * Equality checks:
	 */
	
	/**
	 * Compares SDFs for equality. This is the case if the hashtable entries are identical.
	 * @param o the object this SDF is to be compared to
	 * @return true or false, regarding equality
	 * @see java.lang.Object#equals(java.lang.Object)
	 */
	
	public boolean equals(Object o) {
		if(!(o instanceof SDF) || !this.domain.equals(((SDF)o).domain)) {
			return false;
		}
		SDF p = (SDF)o;
		
		for (SDF_Variable var : p.variables) {
			for (SDF_Variable var2 : p.variables) {
				if(!numericTest(this.values.get(var).get(var2), p.values.get(var).get(var2))) {
					return false;
				}
			}
		}
		return true;
	}
	
	/*
	 * Output methods:
	 */
	
	/**
	 * @return The current SDF as a string.
	 */
	
	public String displayText() {
		Arrays.sort(variables, new Comparator<SDF_Variable>() {
			public int compare(SDF_Variable arg0, SDF_Variable arg1) {
				return arg0.toString().compareToIgnoreCase(arg1.toString());
			}
		});
		
		/*
		 * Determine column width:
		 */
		
		int max = 0;
		for(Variable v : variables) {
			max = (v.toString().length() > max)? v.toString().length() : max;
		}
		max += 3;
		
		String format = "", result = "", line = "";
				
		/*
		 * Prepare format string for title:
		 */
		
		for (int i = 0; i < variables.length+1; i++) {
			format += "|%"+(i+1)+"$-"+max+"s";
		}
		format += "|\n";
		
		/*
		 * Output header line:
		 */
		
		Object[] data = new Object[variables.length+1];
		for (int i = 0; i < variables.length; i++) {
			data[i+1] = variables[i];
		}
		data[0] = "";
		result += String.format(format, data);
		
		/*
		 * Horizontal line:
		 */
		
		for (int i = 0; i < result.length()-1; i++) {
			line += "-";
		}
		line += "\n";
		result = line+result+line;
		
		/*
		 * Prepare format string for configurations:
		 */
		
		format = "|%"+1+"$-"+max+"s";
		for (int i = 1; i < variables.length+1; i++) {
			format += "|%"+(i+1)+"$"+max+"s";	
		}
		format += "|\n";
		
		/*
		 * Configurations:
		 */
		
		for (int i = 0; i < values.size(); i++) {
			Object row[] = new Object[variables.length+1];
			for (int j = 0; j < variables.length; j++) {
				row[j+1] = getSymbol(this.values.get(variables[i]).get(variables[j]));
			}
			row[0] = variables[i];
			result += String.format(format, row);
		}   
		
		result += line;
		
		return result;
	}
	
	/**
	 * Same as displayText(): invokes this method internally.
	 * @see java.lang.Object#toString()
	 */
	
	public String toString() {
		return displayText();
	}
	
	/*
	 * Helper methods:
	 */
	
	private String getSymbol(double d) {
		if(d == Double.POSITIVE_INFINITY) {
			return SYMBOLS.getInfinity(); 
		}
		int rounded = new Double(d).intValue();
		return rounded+"";
	}
	
	/**
	 * Removes the given variable from the current SDF i.e. calculates the marginalization
	 * onto <code>dom = this.dom - {var}</code>. 
	 * @param var The variable to remove.
	 * @return a new, projected SDF object
	 */
	
	private SDF eliminateVariable(SDF_Variable var) {
		//Create new DF_Variable array.
		SDF_Variable[] newVariables = new SDF_Variable[this.variables.length-1];
		int i = 0;
		for (; i < newVariables.length && !this.variables[i].equals(var); i++) {
			newVariables[i] = this.variables[i];
		}
		for (; i < newVariables.length; i++) {
			newVariables[i] = this.variables[i+1];
		}
		
		//Remove the values corresponding to the variable var.
		Hashtable<SDF_Variable, Hashtable<SDF_Variable, Double>> newValues = deepClone(values);
		newValues.remove(var);
		for (SDF_Variable var2 : newVariables) {
			newValues.get(var2).remove(var);
		}
		return new SDF(newVariables, newValues);
	}
	
    /**
     * This test compares two numeric double values on equality.
     * They are equal if their distance is smaller that the EPSILON constant.
     * @param d1 The first value.
     * @param d2 The second value.
     * @return <code>true</code>, if the two values are equal.
     */
    
    protected static boolean numericTest(double d1, double d2) {
        double distance = Math.abs(d1 - d2);
        return !(distance > EPSILON);
    }
}
