package examples;

import gui.TreeViewer;
import indicators.BinaryVariable;
import indicators.Gates;
import indicators.Indicator;

import java.util.Arrays;

import nenok.Architecture;
import nenok.LCFactory;
import nenok.Services;
import nenok.lc.JoinTree;
import nenok.net.Knowledgebase;
import nenok.net.Locator;
import nenok.net.processor.Processor;
import nenok.net.registry.Registry;
import nenok.va.Domain;
import nenok.va.Valuation;

/**
 * This application simulates an n-bit adder with multiple processors running on a singe 
 * machine. The number of processors is specified by the application's first argument. The 
 * second argument - either "Console" or "Interactive" specifies the application mode and finally
 * the fourth argument defines the size of the adder circuit.
 * 
 * @author Marc Pouly
 * @version 1.2
 */

public class Adder {
    
    // The name of the knowledgebase:
    private static final String NAME = "Adder";

    /**
     * @param args Usage: Adder #procs (Interactive | Console) #size
     */
    
    public static void main(String[] args) {
        
        // Verify number of arguments:
        if(args.length != 4) {
            System.out.println("Wrong number of Arguments.");
            System.out.println("Usage: Adder #procs (Interactive | Console) #size (Remote | Local)");
            System.exit(1);
        }
        
        // Fetch number of processors:
        int procs = Integer.parseInt(args[0]);
        
        if(procs < 1) {
            System.out.println("The number of processors should be positive.");
            System.out.println("Usage: Adder #procs (Interactive | Console) #size (Remote | Local)");
            System.exit(1);
        }
        
        // Fetch adder size:
        int size = Integer.parseInt(args[2]);
        
        if(size < 1) {
            System.out.println("The number of processors should be positive.");
            System.out.println("Usage: Adder #procs (Interactive | Console) #size (Remote | Local)");
            System.exit(1);
        }        
         
        /*
         * Variables:
         */

        //Input Variables:
        BinaryVariable[] in = new BinaryVariable[2*size];
        for (int i = 0; i < in.length; i++) {
			in[i] = new BinaryVariable("in"+(i+1));
		}
        
        BinaryVariable inC = new BinaryVariable("inC");

        //Output Variables:
        BinaryVariable[] out = new BinaryVariable[size];
        for (int i = 0; i < out.length; i++) {
        	out[i] = new BinaryVariable("out"+(i+1));
		}
        
        BinaryVariable outC = new BinaryVariable("outC");
        
        /*
         * Indicators:
         */
        
        // Adder circuit:
        Indicator[] adder = Gates.Adder_N(in, inC, out, outC);
        
        // Input lines:
        Indicator[] lines = new Indicator[2*size+1];
        for (int i = 0; i < lines.length - 1; i++) {
			lines[i] = Gates.LINE(in[i], 0);
		}
        
        lines[2*size] =  Gates.LINE(inC, 0);

        //Collect knowledgebase:
        Valuation[] indicators = new Indicator[adder.length+lines.length];
        System.arraycopy(adder, 0, indicators, 0, adder.length);
        System.arraycopy(lines, 0, indicators, adder.length, lines.length);
                                
        /*
         * Queries:
         */
        
        Domain[] queries = new Domain[] {
            new Domain(outC)
        };
                        
        /*
         * Setup NENOK Federation:
         */
        
        if(!args[3].equalsIgnoreCase("Local") && !args[3].equalsIgnoreCase("Remote")) {
            System.out.println("Wrong fourth Argument.");
            System.out.println("Usage: Adder #procs (Interactive | Console) #size (Remote | Local)");
            System.exit(1);
        }
                       
        try {
                       
            // Run Processors :
            Processor[] processors = new Processor[procs]; 
           
            for (int i = 0; i < procs; i++) {
                processors[i] = (args[3].equalsIgnoreCase("Local")) ? Services.startLocalProcessor() : Services.startRemoteProcessor();
            }
            
            Locator[] locators = new Locator[indicators.length]; 
            
            // Store &amp; register potentials:
            for (int i = 0, j = 0; i < indicators.length; i++, j++) {
                locators[i] = processors[j % procs].store(indicators[i]);
            }
            
            Knowledgebase kb = null;
            if(args[3].equalsIgnoreCase("Remote")) {
            	
                // Create knowledgebase in Registry:
                Registry registry = Services.findRegistry();
                registry.createKnowledgebase(NAME);
                registry.addAll(NAME, locators);
                
                kb = registry.getKnowledgebase(NAME);
            	
            } else {
            	
            	// Create local knowledgebase object:
            	kb = new Knowledgebase(locators);
            	kb.setName(NAME);
            	
            }
            
            /*
             * Local Computation:
             */
            
            if(args[1].equalsIgnoreCase("Console")) {
            
                LCFactory factory = new LCFactory();
                factory.setArchitecture(Architecture.Idempotent_Architecture);
                JoinTree tree = factory.create(kb, Arrays.asList(queries), processors[0]);
                tree.propagate();
                
                /*
                 * Query answering:
                 */
                
                System.out.println("Marginals of the Adder Example:\n");
                
                for(Domain query : queries) {
                    System.out.println(tree.answer(query));
                }
                            
            } else if(args[1].equalsIgnoreCase("Interactive")) {
            	            	
            	if(args[3].equalsIgnoreCase("Remote"))
            		TreeViewer.display();
            	TreeViewer.display(kb);
            	                
            } else {
                System.out.println("Wrong second Argument.");
                System.out.println("Usage: Adder #procs (Interactive | Console) #size (Remote | Local)");
            }
            
            /*
             * Destroy Processors:
             */
            
            if(args[3].equalsIgnoreCase("Remote")) {
            	Services.findRegistry().deleteKnowledgebase(NAME);
            }
            
            Services.destroyAllProcessors();

        } catch (Exception e1) {
            e1.printStackTrace();
            System.exit(1);
        }
    }
}
