package indicators;

import java.util.Iterator;

import nenok.va.Domain;
import nenok.va.Idempotency;
import nenok.va.Identity;
import nenok.va.Regularity;
import nenok.va.VAException;
import nenok.va.Valuation;
import nenok.va.Variable;

/**
 * Indicator functions are an example of an idempotent valuation algebra and this
 * class offers a possible implementation of the valuation interface for the case
 * of indicator functions.
 * 
 * @author Marc Pouly
 * @version 1.1
 */

public class Indicator implements Idempotency {
	
    private BinaryVariable[] variables;
    private int[] indicator;
    private Domain domain;
    	
    /*
     * Constructors:
     */
    
	/**
	 * Constructor:
	 * @param variable The single variable of this potential. 
	 * @param indicator The indicator for each possible configuration.<br/> 
	 * The size of the indicator array must be 2.
	 */
	
	public Indicator(BinaryVariable variable, int[] indicator) {
	    if(variable == null || indicator == null) {
	        throw new IllegalArgumentException("Assert arguments being != null.");
	    }
	    setInstanceFields(new BinaryVariable[] {variable}, indicator);
	}
		
	/**
	 * Constructor:
	 * @param variables An array of binary variables.
	 * @param indicator An array of indicators.
	 * The size of the indicator array must be <code>2*variables.length</code>
	 */
	
	public Indicator(BinaryVariable[] variables, int[] indicator) {
	    if(variables == null || indicator == null) {
	        throw new IllegalArgumentException("Assert arguments being != null.");
	    }
	    setInstanceFields(variables, indicator);
	}
		
	/**
	 * Helper method shared by multiple constructors.
	 * Sets instance fields and performs dimension check.
	 * @param variables An array of discrete variables.
	 * @param indicator An array of indicators.
	 */
	
	private void setInstanceFields(BinaryVariable[] variables, int[] indicator) {
	    this.variables = variables;
	    this.indicator = indicator;
	    if(!testDimension() || !testContent()) {
	        throw new IllegalArgumentException("Dimension of indicator array doesn't match. Given: "+indicator.length+", expected: "+countConfigurations()+".");
	    }
	}

    /*
     * Valuation implementation:
     */
    
    /**
     * @see nenok.va.Valuation#label()
     */
    
	@SuppressWarnings("cast") 
    public Domain label() {
    	if(domain == null) {
    		domain = new Domain((Variable[])variables);
    	}
        return domain;
    }

    /**
     * @see nenok.va.Valuation#combine(nenok.va.Valuation)
     */
      
    public Valuation combine(Valuation val) {
		if(val instanceof Identity) {
			return this;
		}
        Indicator arg = (Indicator)val;
        
        //Number of variable in the new domain:
        int inter = 0;
        for (int i = 0; i < variables.length; i++) {
			if(indexOf(variables[i], arg.variables) >= 0) {
				inter++;
			}
		}
        
		//Variable array of the new indicator function:
		BinaryVariable[] newVariables = new BinaryVariable[arg.variables.length + variables.length - inter];        
		Variable[] interVars = new Variable[inter];
        
		//Fill new variable array and compute dimension of indicator array
		int interCounter = 0;
		int indicatorSize = (int)Math.pow(2, newVariables.length); 
		System.arraycopy(variables, 0, newVariables, 0, variables.length);
		for (int j = 0, counter = variables.length; j < arg.variables.length; j++) {
			if(indexOf(arg.variables[j], variables) < 0) {
				newVariables[counter++] = arg.variables[j];
			} else {
				interVars[interCounter++] = arg.variables[j];
			}
		}
		
		//Indicator array of the new potential:
		int[] newIndicator = new int[indicatorSize];
		
		/*
		 * Compute each configuration's value
		 */
			
		int[] symbols = new int[inter];
		int thisConfgurations = countConfigurations();
		int argConfigurations = arg.countConfigurations();
		
		// Determine indexes of intervars:
		int thisIndex[] = new int[symbols.length];
		int argIndex[] = new int[symbols.length];
		for (int i = 0; i < interVars.length; i++) {
			thisIndex[i] = indexOf(interVars[i], variables);
			argIndex[i] = indexOf(interVars[i], arg.variables);
		}
		
		for (int i = 0, configurationIndex = 0; i < thisConfgurations; i++) {
		    //Store all variables' indexes of this.variables:
		    for (int j = 0; j < interVars.length; j++) {
		    	symbols[j] = getSymbolIndex(thisIndex[j], i);
            }
		    for (int j = 0; j < argConfigurations; j++) {
		        boolean multiply = true;
		        for (int k = 0; k < interVars.length; k++) {
		        	multiply &= (arg.getSymbolIndex(argIndex[k], j) == symbols[k]);
                }
		        //Perform multiplication (corresponds to logical AND):
		        if(multiply) {
		            newIndicator[configurationIndex++] = this.indicator[i] * arg.indicator[j];
		        }
            }
        }
		return new Indicator(newVariables, newIndicator);
    }

    /**
     * @see nenok.va.Valuation#marginalize(nenok.va.Domain)
     */
    
    
	public Valuation marginalize(Domain dom) throws VAException {
		if(label().equals(dom)) {
		    return this;
		}
		if(!dom.subSetOf(label())) {
			throw new VAException("Marginalization impossible: this.label() = "+label()+", argument = "+dom+".");
		}
		Indicator result = this;
		//Calculates the difference set between both domains.
		Domain difference = Domain.difference(label(), dom);
		Iterator it = difference.iterator();
		//Remove each variable in the difference set.
		while(it.hasNext()) {
			BinaryVariable var = (BinaryVariable)it.next();
			result = result.eliminateVariable(var);
		}
		return result;
	}
	    
    /**
     * @see nenok.va.Valuation#weight()
     */
    
    
    public int weight() {
        return (int)Math.pow(2,variables.length);
    }
    
    /*
     * Dividability implementation:
     */
    
    /**
     * @see nenok.va.Regularity#inverse()
     */
    
    
    public Regularity inverse() {
        return new Idempotency.Implementor().inverse(this);
    }
    		    	
	/*
	 * Equality check:
	 */
	
    /**
     * Equality test:
     * Inequality is certain. Equality is based on heuristic for performance reasons.
     * @see java.lang.Object#equals(java.lang.Object)
     */
	
	public boolean equals(Object o) {
		if(!(o instanceof Indicator)) {
			return false;
		}
		if(this == o) {
			return true;
		}
		Indicator p = (Indicator)o;
		if(variables.length != p.variables.length || indicator.length != p.indicator.length) {
			return false;
		}
		
		// Domain equality:
		for(Variable v : variables) {
			if(indexOf(v, p.variables) < 0) {
				return false;
			}
		}
		
		for (int i = 0; i < p.variables.length; i++) {
			if(indexOf(p.variables[i], variables) < 0) {
				return false;
			}
		}
		
		int sum1 = 0, sum2 = 0;
		for (int i = 0; i < indicator.length; i++) {
			sum1 += indicator[i];
			sum2 += p.indicator[i];
		}
		
		return sum1 == sum2;		
	}
			
    /*
     * Output methods:
     */
    
    /**
     * @return The current indicator function as a string. 
     * Displays all configurations..
     */
    
    public String displayFullData() {
        int max = 0;
        String format = "", result = "", line = "", title = "Ind.";
        
        /*
         * Find string of maximum length:
         */
        
        for (int i = 0; i < variables.length; i++) {
            BinaryVariable v = variables[i];
            if(max < v.toString().length()) {
                max = v.toString().length();
            }
        }
        
        if(max < title.length()) {
            max = title.length();
        }
        
        max += 2;
        
        /*
         * Prepare format string for title:
         */

        for (int i = 0; i < variables.length+1; i++) {
            format += "|%"+(i+1)+"$-"+max+"s";
        }
        format += "|\n";
        
        /*
         * Output header line:
         */
        
        Object[] data = new Object[variables.length+1];
        for (int i = 0; i < variables.length; i++) {
            data[i] = variables[i];
        }
        data[variables.length] = title;
        result += String.format(format, data);
        
        /*
         * Horizontal line:
         */
        
        for (int i = 0; i < result.length()-1; i++) {
            line += "-";
        }
        line += "\n";
        result = line+result+line;
        
        /*
         * Configurations:
         */
        
        for (int i = 0; i < indicator.length; i++) {
            Object row[] = new Object[data.length];
            for (int j = 0; j < variables.length; j++) {
                row[j] = variables[j].getFrame()[getSymbolIndex(j,i)].toString();
            }
            row[data.length-1] = indicator[i];
            result += String.format(format, row);
        }   
        
        result += line;
        return result;
    }
       
    /**
     * @return The current indicator function as a string. 
     * Displays only configurations with indicator == 1.
     */
    
    public String displayRelevantData() {
        int max = 0;
        String format = "", result = "", line = "", title = "Ind.";
        
        /*
         * Find string of maximum length:
         */
        
        for (int i = 0; i < variables.length; i++) {
            BinaryVariable v = variables[i];
            if(max < v.toString().length()) {
                max = v.toString().length();
            }
        }
        
        if(max < title.length()) {
            max = title.length();
        }
        
        max += 2;
        
        /*
         * Prepare format string for title:
         */

        for (int i = 0; i < variables.length+1; i++) {
            format += "|%"+(i+1)+"$-"+max+"s";
        }
        format += "|\n";
        
        /*
         * Output header line:
         */
        
        Object[] data = new Object[variables.length+1];
        for (int i = 0; i < variables.length; i++) {
            data[i] = variables[i];
        }
        data[variables.length] = title;
        result += String.format(format, data);
        
        /*
         * Horizontal line:
         */
        
        for (int i = 0; i < result.length()-1; i++) {
            line += "-";
        }
        line += "\n";
        result = line+result+line;
        
        /*
         * Configurations:
         */
        
        for (int i = 0; i < indicator.length; i++) {
            if(indicator[i] == 1) {
                Object row[] = new Object[data.length];
                for (int j = 0; j < variables.length; j++) {
                    row[j] = variables[j].getFrame()[getSymbolIndex(j,i)].toString();
                }
                row[data.length-1] = indicator[i];
                result += String.format(format, row);   
            }
        }   
        
        result += line;
        return result;
    }
    
    /**
     * @see java.lang.Object#toString()
     */
    
    public String toString() {
        return displayRelevantData();
    }
    	    
    /*
     * Helper methods:
     */
    
    /**
     * Removes the given variable from the current indicator function 
     * i.e. computes the marginalization onto <code>dom = this.dom - {var}</code>. 
     * Generates a new indicator function instance - the original indicator function 
     * stays untouched.
     * @param var The variable to remove.
     * @return The resulting indicator function.
     */
    
    private Indicator eliminateVariable(BinaryVariable var) {
        //The new variable array.
        BinaryVariable[] newVars = new BinaryVariable[this.variables.length - 1];
        int oldPosition = -1;
        int size = 1;
        //Fill the new variable array and calculate the number of configurations of the new potential.
        for (int i = 0, j = 0; i < this.variables.length; i++) {
            if(!this.variables[i].equals(var)) {
                size *= 2;
                newVars[j++] = this.variables[i];
            } else {
                oldPosition = i;
            }
        }
        //The new indicator array.
        int[] newIndicator = new int[size];
        int offset = (int)Math.pow(2,variables.length - oldPosition - 1);
        //Calculate the elements of the new probability array.
        int blocs = (indicator.length / 2) / offset;
        int blocLines = newIndicator.length / blocs;
        for (int i = 0, index = 0; i < blocs; i++) {
            int base = i * (indicator.length / blocs);
            for (int j = 0; j < blocLines; j++) {
                int sum = 0;                
                for (int k = 0; k < 2; k++) {   
                    sum = sum + this.indicator[base + j + k*offset];
                    if(sum > 1) {
                        sum = 1;
                    }
                }
                newIndicator[index] = sum;
                index++;
            }
        }
        //Generate and return the new instance.
        return new Indicator(newVars, newIndicator);
    }
    
    /**
     * This method returns the index of the variable's value at a specifique configuration.
     * @param varPos The variable caracterized by its position.
     * @param configuration The configuration
     * @return This symbol of the variable at the given configuration.
     * Example:<br/>
     * <code>getSymbol(0,3)</code> returns the index of variable 0's symbol at configuration 3.
     */
    
    private int getSymbolIndex(int varPos, int configuration) {
        int repetitions = (int)Math.pow(2,variables.length - varPos - 1);
        return (configuration / repetitions) % 2;
    }
    
    /**
     * @return Does the current indicator array's dimension 
     * correspond to the number of configurations.
     */
    
    private boolean testDimension() {
        return countConfigurations() == indicator.length;
    }
    
    /**
     * @return Does the current indicator array only contain values from set {0,1}.
     */
    
    private boolean testContent() {
	    for (int i = 0; i < indicator.length; i++) {
            if(indicator[i] != 0 && indicator[i] != 1) {
                throw new IllegalArgumentException("Indicator values must be from set {0,1}.");                
            }
        }
	    return true;
    }
    
    /**
     * @return Compute the number of configurations.
     */
    
    private int countConfigurations() {
    	return (int)Math.pow(2,variables.length);      
    }
    
    /**
     * Returns the index of the given variable in the variable array.
     * @param variable The variable whose index is searched.
     * @param variables The variable array which contains the index.
     * @return the variable's index or <code>-1</code>.
     */
    
    private static int indexOf(Variable variable, Variable[] variables) {
        for (int i = 0; i < variables.length; i++) {
            if(variables[i].equals(variable)) {
                return i;
            }
        }
        return -1;
    }
}