package examples;

import gui.TreeViewer;

import java.util.Arrays;

import nenok.Architecture;
import nenok.LCFactory;
import nenok.Services;
import nenok.lc.JoinTree;
import nenok.net.Knowledgebase;
import nenok.net.Locator;
import nenok.net.processor.Processor;
import nenok.net.registry.Registry;
import nenok.va.Domain;
import nenok.va.Valuation;
import ppotentials.PP_Variable;
import ppotentials.Potential;
import ppotentials.SRPotential;

/**
 * This application class solves the Asia example with multiple processors running on a singe machine. The number of processors 
 * is specified by the application's first argument. With the second argument, the user can choose between two implementations. 
 * The flag "Semiring" indicates that semiring induced probability potentials are used whereas "Normal" stresses the default 
 * implementation. The third argument - either "Console" or "Interactive" specifies the application mode and finally, the fourth
 * argument indicates either "Local" or "Remote" execution.
 * 
 * @author Marc Pouly
 * @version 1.2
 */

public class Asia {
    
    // The name of the knowledgebase:
    private static final String NAME = "Asia";

    /**
     * @param args Usage: Asia int (Normal | Semiring) (Interactive | Console) (Local | Remote)
     */
    
    public static void main(String[] args) {
        
        // Verify number of arguments:
        if(args.length != 4) {
            System.out.println("Wrong number of Arguments.");
            System.out.println("Usage: Asia int (Normal | Semiring) (Interactive | Console) (Local | Remote)");
            System.exit(1);
        }
        
        // Fetch number of processors:
        int procs = Integer.parseInt(args[0]);
        if(procs < 1) {
            System.out.println("The number of processors should be positive.");
            System.out.println("Usage: Asia int (Normal | Semiring) (Interactive | Console) (Local | Remote)");
            System.exit(1);
        }
        
        Valuation[] potentials = null;
        
        /*
         * Frame:
         */
        
        //There are only binary variables:
        String[] frame = { "0", "1" };
                
        /*
         * Variables:
         */

        //Visit to Asia:
        PP_Variable a = new PP_Variable("A", frame);
        //Tuberculosis:
        PP_Variable t = new PP_Variable("T", frame);
        //Smoking:
        PP_Variable s = new PP_Variable("S", frame);
        //Bronchitis:
        PP_Variable b = new PP_Variable("B", frame);
        //Lung cancer:
        PP_Variable l = new PP_Variable("L", frame);
        //Either tubercolosis or lung cancer:
        PP_Variable e = new PP_Variable("E", frame);
        //Positive X-ray:
        PP_Variable x = new PP_Variable("X", frame);
        //Dyspnoea:
        PP_Variable d = new PP_Variable("D", frame);
        
        /*
         * Default implementation:
         */
        
        if(args[1].equalsIgnoreCase("Normal")) {
            
            double[] probs1 = {0.99, 0.01};
            Potential pot1 = new Potential(a, probs1);
            
            PP_Variable[] cond1 = {a};
            double[] probs2 = {0.99, 0.95, 0.01, 0.05};
            Potential pot2 = new Potential(t, cond1, probs2);
            
            double[] probs3 = {0.5, 0.5};
            Potential pot3 = new Potential(s, probs3);
            
            PP_Variable[] cond2 = {s};
            double[] probs4 = {0.99, 0.9, 0.01, 0.1};
            Potential pot4 = new Potential(l, cond2, probs4);
            
            PP_Variable[] cond3 = {s};
            double[] probs5 = {0.7, 0.4, 0.3, 0.6};
            Potential pot5 = new Potential(b, cond3, probs5);
            
            PP_Variable[] cond4 = {l, t};
            double[] probs6 = {1, 0, 0, 0, 0, 1, 1, 1};
            Potential pot6 = new Potential(e, cond4, probs6);
            
            PP_Variable[] cond5 = {e};
            double[] probs7 = {0.95, 0.02, 0.05, 0.98};
            Potential pot7 = new Potential(x, cond5, probs7);   
            
            PP_Variable[] cond6 = {e, b};
            double[] probs8 = {0.9, 0.2, 0.3, 0.1, 0.1, 0.8, 0.7, 0.9};
            Potential pot8 = new Potential(d, cond6, probs8);
            
            potentials = new Potential[] {pot1, pot2, pot3, pot4, pot5, pot6, pot7, pot8};
        } 
        
        /*
         * Semiring implementation:
         */
        
        else if(args[1].equalsIgnoreCase("Semiring")) {
            
            double[] probs1 = {0.99, 0.01};
            SRPotential pot1 = new SRPotential(a, probs1);
            
            PP_Variable[] cond1 = {a};
            double[] probs2 = {0.99, 0.95, 0.01, 0.05};
            SRPotential pot2 = new SRPotential(t, cond1, probs2);
            
            double[] probs3 = {0.5, 0.5};
            SRPotential pot3 = new SRPotential(s, probs3);
            
            PP_Variable[] cond2 = {s};
            double[] probs4 = {0.99, 0.9, 0.01, 0.1};
            SRPotential pot4 = new SRPotential(l, cond2, probs4);
            
            PP_Variable[] cond3 = {s};
            double[] probs5 = {0.7, 0.4, 0.3, 0.6};
            SRPotential pot5 = new SRPotential(b, cond3, probs5);
            
            PP_Variable[] cond4 = {l, t};
            double[] probs6 = {1, 0, 0, 0, 0, 1, 1, 1};
            SRPotential pot6 = new SRPotential(e, cond4, probs6);
            
            PP_Variable[] cond5 = {e};
            double[] probs7 = {0.95, 0.02, 0.05, 0.98};
            SRPotential pot7 = new SRPotential(x, cond5, probs7);   
            
            PP_Variable[] cond6 = {e, b};
            double[] probs8 = {0.9, 0.2, 0.3, 0.1, 0.1, 0.8, 0.7, 0.9};
            SRPotential pot8 = new SRPotential(d, cond6, probs8);
            
            potentials = new SRPotential[] {pot1, pot2, pot3, pot4, pot5, pot6, pot7, pot8};
        } 
        
        else {
            
            System.out.println("Wrong second Argument.");
            System.out.println("Usage: Asia int (Normal | Semiring) (Interactive | Console) (Local | Remote)");
            System.exit(1);
            
        }
        
        /*
         * Queries:
         */
        
        Domain[] queries = new Domain[] {
            new Domain(a),
            new Domain(t),
            new Domain(s),
            new Domain(b),
            new Domain(l),
            new Domain(x),
            new Domain(e),
            new Domain(d)
        };
        
        if(!args[3].equalsIgnoreCase("Local") && !args[3].equalsIgnoreCase("Remote")) {
            System.out.println("Wrong fourth Argument.");
            System.out.println("Usage: Asia int (Normal | Semiring) (Interactive | Console) (Local | Remote)");
            System.exit(1);
        }
                       
        /*
         * Setup NENOK Federation:
         */
              
        try {
                       
            // Run Processors :
            Processor[] processors = new Processor[procs]; 
           
            for (int i = 0; i < procs; i++) {
                processors[i] = (args[3].equalsIgnoreCase("Local")) ? Services.startLocalProcessor() : Services.startRemoteProcessor();
            }
            
            Locator[] locators = new Locator[potentials.length]; 
            
            // Store &amp; register potentials:
            for (int i = 0, j = 0; i < potentials.length; i++, j++) {
                locators[i] = processors[j % procs].store(potentials[i]);
            }
            
            Knowledgebase kb = null;
            if(args[3].equalsIgnoreCase("Remote")) {
            	
                // Create knowledgebase in Registry:
                Registry registry = Services.findRegistry();
                registry.createKnowledgebase(NAME);
                registry.addAll(NAME, locators);
                
                kb = registry.getKnowledgebase(NAME);
            	
            } else {
            	
            	// Create local knowledgebase object:
            	kb = new Knowledgebase(locators);
            	kb.setName(NAME);
            	
            }
            
            /*
             * Local Computation:
             */
            
            if(args[2].equalsIgnoreCase("Console")) {
            
                LCFactory factory = new LCFactory();
                factory.setArchitecture(Architecture.Lauritzen_Spiegelhalter);
                JoinTree tree = factory.create(kb, Arrays.asList(queries), processors[0]);
                tree.propagate();
                
                /*
                 * Query answering:
                 */
                
                System.out.println("Marginals of the Asia Example:\n");
                
                for(Domain query : queries) {
                    System.out.println(tree.answer(query));
                }
                            
            } else if(args[2].equalsIgnoreCase("Interactive")) {
            	            	
            	if(args[3].equalsIgnoreCase("Remote"))
            		TreeViewer.display();
            	TreeViewer.display(kb);
                
            } else {
                System.out.println("Wrong third Argument.");
                System.out.println("Usage: Asia int (Normal | Semiring) (Interactive | Console) (Local | Remote)");
            }
            
            /*
             * Destroy Processors:
             */
            
            if(args[3].equalsIgnoreCase("Remote")) {
            	Services.findRegistry().deleteKnowledgebase(NAME);
            }
            
            Services.destroyAllProcessors();

        } catch (Exception e1) {
            e1.printStackTrace();
            System.exit(1);
        }
    }
}
