package examples;

import gui.TreeViewer;

import java.util.Arrays;

import nenok.Architecture;
import nenok.LCFactory;
import nenok.Services;
import nenok.lc.JoinTree;
import nenok.net.Knowledgebase;
import nenok.net.Locator;
import nenok.net.processor.Processor;
import nenok.net.registry.Registry;
import nenok.va.Domain;
import nenok.va.Valuation;
import ppotentials.PP_Variable;
import ppotentials.Potential;
import ppotentials.SRPotential;

/**
 * This application class solves the Earthquake example with multiple processors running on a singe machine. The number of 
 * processors is specified by the application's first argument. With the second argument, the user can choose between two 
 * implementations. The flag "Semiring" indicates that semiring induced probability potentials are used whereas "Normal" stresses 
 * the default implementation. The third argument - either "Console" or "Interactive" specifies the application mode and finally, 
 * the fourth argument indicates either "Local" or "Remote" execution.
 * 
 * @author Marc Pouly
 * @version 1.2
 */

public class Earthquake {
    
    // The name of the knowledgebase:
    private static final String NAME = "Earthquake";

    /**
     * @param args
     */
    
    public static void main(String[] args) {
        
        // Verify number of arguments:
        if(args.length != 4) {
            System.out.println("Wrong number of Arguments.");
            System.out.println("Usage: Earthquake int (Normal | Semiring) (Interactive | Console) (Local | Remote)");
            System.exit(1);
        }
        
        // Fetch number of processors:
        int procs = Integer.parseInt(args[0]);
        if(procs < 1) {
            System.out.println("The number of processors should be positive.");
            System.out.println("Usage: Earthquake int (Normal | Semiring) (Interactive | Console) (Local | Remote)");
            System.exit(1);
        }
        
        Valuation[] potentials = null;
        
        /*
         * Frame:
         */
        
        // There are only binary variables:
        String[] frame = { "0", "1" };
                
        /*
         * Variables:
         */
        
        // Burglary
        PP_Variable b = new PP_Variable("B", frame);
        // Earthquake
        PP_Variable e = new PP_Variable("E", frame);
        // John calls:
        PP_Variable j = new PP_Variable("J", frame);
        // Mary calls:
        PP_Variable m = new PP_Variable("M", frame);
        // Alarm
        PP_Variable a = new PP_Variable("A", frame);
        
        /*
         * Default implementation:
         */
        
        if(args[1].equalsIgnoreCase("Normal")) {
            
            double[] probs1 = {0.999, 0.001};
            Potential pot1 = new Potential(b, probs1);
            
            double[] probs2 = {0.998, 0.002};
            Potential pot2 = new Potential(e, probs2);   
            
            PP_Variable[] cond1 = {a};
            double[] probs3 = {0.95, 0.1, 0.05, 0.9};
            Potential pot3 = new Potential(j, cond1, probs3);
            
            double[] probs4 = {0.99, 0.3, 0.01, 0.7};
            Potential pot4 = new Potential(m, cond1, probs4);   
            
            PP_Variable[] cond2 = {b, e};
            double[] probs5 = {0.999, 0.71, 0.06,  0.05, 0.001, 0.29, 0.94,  0.95};
            Potential pot5 = new Potential(a, cond2, probs5);
            
            potentials = new Potential[] {pot1, pot2, pot3, pot4, pot5};
        } 
                    
        /*
         * Semiring implementation:
         */
        
        else if(args[1].equalsIgnoreCase("Semiring")) {
        	
            double[] probs1 = {0.999, 0.001};
            SRPotential pot1 = new SRPotential(b, probs1);
            
            double[] probs2 = {0.998, 0.002};
            SRPotential pot2 = new SRPotential(e, probs2);            
            
            PP_Variable[] cond1 = {a};

            double[] probs3 = {0.95, 0.1, 0.05, 0.9};
            SRPotential pot3 = new SRPotential(j, cond1, probs3);
            
            double[] probs4 = {0.99, 0.3, 0.01, 0.7};
            SRPotential pot4 = new SRPotential(m, cond1, probs4);            
            
            PP_Variable[] cond2 = {b, e};
            double[] probs5 = {0.999, 0.71, 0.06, 0.05, 0.001, 0.29, 0.94, 0.95};
            SRPotential pot5 = new SRPotential(a, cond2, probs5);
            
            potentials = new SRPotential[] {pot1, pot2, pot3, pot4, pot5};
        } 
        
        else {
            
            System.out.println("Wrong second Argument.");
            System.out.println("Usage: Earthquake int (Normal | Semiring) (Interactive | Console) (Local | Remote)");
            System.exit(1);
            
        }
            
        /*
         * Queries:
         */
        
        Domain[] queries = new Domain[] {
            new Domain(m),
            new Domain(a),
            new Domain(e),
            new Domain(b),
            new Domain(j)
        };
                                
        if(!args[3].equalsIgnoreCase("Local") && !args[3].equalsIgnoreCase("Remote")) {
            System.out.println("Wrong fourth Argument.");
            System.out.println("Usage: Earthquake int (Normal | Semiring) (Interactive | Console) (Local | Remote)");
            System.exit(1);
        }
                       
        /*
         * Setup NENOK Federation:
         */
              
        try {
                       
            // Run Processors :
            Processor[] processors = new Processor[procs]; 
           
            for (int i = 0; i < procs; i++) {
                processors[i] = (args[3].equalsIgnoreCase("Local")) ? Services.startLocalProcessor() : Services.startRemoteProcessor();
            }
            
            Locator[] locators = new Locator[potentials.length]; 
            
            // Store &amp; register potentials:
            for (int i = 0, k = 0; i < potentials.length; i++, k++) {
                locators[i] = processors[k % procs].store(potentials[i]);
            }
            
            Knowledgebase kb = null;
            if(args[3].equalsIgnoreCase("Remote")) {
            	
                // Create knowledgebase in Registry:
                Registry registry = Services.findRegistry();
                registry.createKnowledgebase(NAME);
                registry.addAll(NAME, locators);
                
                kb = registry.getKnowledgebase(NAME);
            	
            } else {
            	
            	// Create local knowledgebase object:
            	kb = new Knowledgebase(locators);
            	kb.setName(NAME);
            	
            }
            
            /*
             * Local Computation:
             */
            
            if(args[2].equalsIgnoreCase("Console")) {
            
                LCFactory factory = new LCFactory();
                factory.setArchitecture(Architecture.Lauritzen_Spiegelhalter);
                JoinTree tree = factory.create(kb, Arrays.asList(queries), processors[0]);
                tree.propagate();
                
                /*
                 * Query answering:
                 */
                
                System.out.println("Marginals of the Earthquake Example:\n");
                
                for(Domain query : queries) {
                    System.out.println(tree.answer(query));
                }
                            
            } else if(args[2].equalsIgnoreCase("Interactive")) {
                
            	if(args[3].equalsIgnoreCase("Remote"))
            		TreeViewer.display();
            	TreeViewer.display(kb);
                
            } else {
                System.out.println("Wrong third Argument.");
                System.out.println("Usage: Earthquake int (Normal | Semiring) (Interactive | Console) (Local | Remote)");
            }
            
            /*
             * Destroy Processors:
             */
            
            if(args[3].equalsIgnoreCase("Remote")) {
            	Services.findRegistry().deleteKnowledgebase(NAME);
            }
            
            Services.destroyAllProcessors();

        } catch (Exception e1) {
            e1.printStackTrace();
            System.exit(1);
        }
    }
}
