package examples;

import gui.TreeViewer;

import java.util.Arrays;

import nenok.Architecture;
import nenok.LCFactory;
import nenok.Services;
import nenok.lc.JoinTree;
import nenok.net.Knowledgebase;
import nenok.net.Locator;
import nenok.net.processor.Processor;
import nenok.net.registry.Registry;
import nenok.va.Domain;
import nenok.va.Valuation;
import ppotentials.PP_Variable;
import ppotentials.Potential;
import ppotentials.SRPotential;

/**
 * This application class solves the Oil Windcatter example with multiple processors running on a singe machine. The number 
 * of processors is specified by the application's first argument. With the second argument, the user can choose between two 
 * implementations. The flag "Semiring" indicates that semiring induced probability potentials are used whereas "Normal" 
 * stresses the default implementation. The third argument - either "Console" or "Interactive" specifies the application mode 
 * and finally, the fourth argument indicates either "Local" or "Remote" execution.
 * 
 * @author Marc Pouly
 * @version 1.2
 */

public class Oil {
    
    // The name of the knowledgebase:
    private static final String NAME = "Oil";

    /**
     * @param args
     */
    
    public static void main(String[] args) {
        
        // Verify number of arguments:
        if(args.length != 4) {
            System.out.println("Wrong number of Arguments.");
            System.out.println("Usage: Oil int (Normal | Semiring) (Interactive | Console) (Local | Remote)");
            System.exit(1);
        }
        
        // Fetch number of processors:
        int procs = Integer.parseInt(args[0]);
        if(procs < 1) {
            System.out.println("The number of processors should be positive.");
            System.out.println("Usage: Oil int (Normal | Semiring) (Interactive | Console) (Local | Remote)");
            System.exit(1);
        }
        
        Valuation[] potentials = null;
                        
        /*
         * Variables:
         */

        PP_Variable oil = new PP_Variable("Oil", new String[] {"dry", "wet", "soak"});
        PP_Variable test = new PP_Variable("Test", new String[] {"yes", "not"});
        PP_Variable seismic = new PP_Variable("Seismic", new String[] {"closed", "open", "diff"});
        
        /*
         * Default implementation:
         */
        
        if(args[1].equalsIgnoreCase("Normal")) {
            
            double[] prob1 = new double[] {0.5, 0.3, 0.2};
            Potential p1 = new Potential(oil, prob1);
            
            double[] prob2 = new double[] {0.33, 0.1, 0.33, 0.3, 0.33, 0.5, 0.33, 0.3, 
                                           0.33, 0.4, 0.33, 0.4, 0.33, 0.6, 0.33, 0.3, 0.33, 0.1};
            Potential p2 = new Potential(seismic, new PP_Variable[] {oil, test}, prob2);        
            
            potentials = new Potential[] {p1, p2};
        } 
        
        /*
         * Semiring implementation:
         */
        
        else if(args[1].equalsIgnoreCase("Semiring")) {
            
            double[] prob1 = new double[] {0.5, 0.3, 0.2};
            SRPotential p1 = new SRPotential(oil, prob1);
            
            double[] prob2 = new double[] {0.33, 0.1, 0.33, 0.3, 0.33, 0.5, 0.33, 0.3, 
                                           0.33, 0.4, 0.33, 0.4, 0.33, 0.6, 0.33, 0.3, 0.33, 0.1};
            SRPotential p2 = new SRPotential(seismic, new PP_Variable[] {oil, test}, prob2);        
            
            potentials = new SRPotential[] {p1, p2};
        } 
        
        else {
            
            System.out.println("Wrong second Argument.");
            System.out.println("Usage: Oil int (Normal | Semiring) (Interactive | Console) (Local | Remote)");
            System.exit(1);
            
        }
        
        /*
         * Queries:
         */
        
        Domain[] queries = new Domain[] {
            new Domain(oil),
            new Domain(test),
            new Domain(seismic)
        };
                
        if(!args[3].equalsIgnoreCase("Local") && !args[3].equalsIgnoreCase("Remote")) {
            System.out.println("Wrong fourth Argument.");
            System.out.println("Usage: Oil int (Normal | Semiring) (Interactive | Console) (Local | Remote)");
            System.exit(1);
        }
                       
        /*
         * Setup NENOK Federation:
         */
              
        try {
                       
            // Run Processors :
            Processor[] processors = new Processor[procs]; 
           
            for (int i = 0; i < procs; i++) {
                processors[i] = (args[3].equalsIgnoreCase("Local")) ? Services.startLocalProcessor() : Services.startRemoteProcessor();
            }
            
            Locator[] locators = new Locator[potentials.length]; 
            
            // Store &amp; register potentials:
            for (int i = 0, j = 0; i < potentials.length; i++, j++) {
                locators[i] = processors[j % procs].store(potentials[i]);
            }
            
            Knowledgebase kb = null;
            if(args[3].equalsIgnoreCase("Remote")) {
            	
                // Create knowledgebase in Registry:
                Registry registry = Services.findRegistry();
                registry.createKnowledgebase(NAME);
                registry.addAll(NAME, locators);
                
                kb = registry.getKnowledgebase(NAME);
            	
            } else {
            	
            	// Create local knowledgebase object:
            	kb = new Knowledgebase(locators);
            	kb.setName(NAME);
            	
            }
            
            /*
             * Local Computation:
             */
            
            if(args[2].equalsIgnoreCase("Console")) {
            
                LCFactory factory = new LCFactory();
                factory.setArchitecture(Architecture.Lauritzen_Spiegelhalter);
                JoinTree tree = factory.create(kb, Arrays.asList(queries), processors[0]);
                tree.propagate();
                
                /*
                 * Query answering:
                 */
                
                System.out.println("Marginals of the Oil Example:\n");
                
                for(Domain query : queries) {
                    System.out.println(tree.answer(query));
                }
                            
            } else if(args[2].equalsIgnoreCase("Interactive")) {
                
            	if(args[3].equalsIgnoreCase("Remote"))
            		TreeViewer.display();
            	TreeViewer.display(kb);
                
            } else {
                System.out.println("Wrong third Argument.");
                System.out.println("Usage: Oil int (Normal | Semiring) (Interactive | Console) (Local | Remote)");
            }
            
            /*
             * Destroy Processors:
             */
            
            if(args[3].equalsIgnoreCase("Remote")) {
            	Services.findRegistry().deleteKnowledgebase(NAME);
            }
            
            Services.destroyAllProcessors();

        } catch (Exception e1) {
            e1.printStackTrace();
            System.exit(1);
        }
    }
}
