package examples;

import gui.TreeViewer;

import java.util.Arrays;

import nenok.Architecture;
import nenok.LCFactory;
import nenok.Services;
import nenok.lc.JoinTree;
import nenok.net.Knowledgebase;
import nenok.net.Locator;
import nenok.net.processor.Processor;
import nenok.net.registry.Registry;
import nenok.va.Domain;
import nenok.va.Valuation;
import ppotentials.PP_Variable;
import ppotentials.Potential;
import ppotentials.SRPotential;

/**
 * This application class solves the Studfarm example with multiple processors running on a singe machine. The number of 
 * processors is specified by the application's first argument. With the second argument, the user can choose between two 
 * implementations. The flag "Semiring" indicates that semiring induced probability potentials are used whereas "Normal" 
 * stresses the default implementation. The third argument - either "Console" or "Interactive" specifies the application mode 
 * and finally, the fourth argument indicates either "Local" or "Remote" execution.
 * 
 * @author Marc Pouly
 * @version 1.2
 */

public class StudFarm {
    
    // The name of the knowledgebase:
    private static final String NAME = "Studfarm";

    /**
     * @param args
     */
    
    public static void main(String[] args) {
        
        // Verify number of arguments:
        if(args.length != 4) {
            System.out.println("Wrong number of Arguments.");
            System.out.println("Usage: Studfarm int (Normal | Semiring) (Interactive | Console) (Local | Remote)");
            System.exit(1);
        }
        
        // Fetch number of processors:
        int procs = Integer.parseInt(args[0]);
        if(procs < 1) {
            System.out.println("The number of processors should be positive.");
            System.out.println("Usage: Studfarm int (Normal | Semiring) (Interactive | Console) (Local | Remote)");
            System.exit(1);
        }
        
        Valuation[] potentials = null;
        
        /*
         * Frames:
         */
        
        String[] frame1 = {"AA", "Aa"};
        String[] frame2 = {"AA", "Aa", "aa"};
        
        /*
         * Variables:
         */
        
        PP_Variable a1 = new PP_Variable("A1", frame1);
        PP_Variable a2 = new PP_Variable("A2", frame1);
        
        PP_Variable ann = new PP_Variable("Ann", frame1);
        PP_Variable alan = new PP_Variable("Alan", frame1);
        PP_Variable alice = new PP_Variable("Alice", frame1);
        
        PP_Variable bill = new PP_Variable("Bill", frame1);
        PP_Variable betsy = new PP_Variable("Betsy", frame1);
        PP_Variable benny = new PP_Variable("Benny", frame1);
        PP_Variable bonnie = new PP_Variable("Bonnie", frame1);
        
        PP_Variable carl = new PP_Variable("Carl", frame1);
        PP_Variable cecily = new PP_Variable("Cecily", frame1);
        
        PP_Variable dennis = new PP_Variable("Dennis", frame2);
        
        /*
         * Default implementation:
         */
        
        if(args[1].equalsIgnoreCase("Normal")) {
            
            //Level 1:
            double[] p1 = {0.99, 0.01};
            Potential p11 = new Potential(a1, p1);
            Potential p12 = new Potential(ann, p1);
            Potential p13 = new Potential(alan, p1);
            Potential p14 = new Potential(alice, p1);
            Potential p15 = new Potential(a2, p1);
            
            //Level 2:
            double[] p2 = {1, 0.5, 0.5, 0.33, 0, 0.5, 0.5, 0.67};
            Potential p21 = new Potential(bill, new PP_Variable[] {a1,ann}, p2);
            Potential p22 = new Potential(betsy, new PP_Variable[] {ann,alan}, p2);
            Potential p23 = new Potential(benny, new PP_Variable[] {alan,alice}, p2);
            Potential p24 = new Potential(bonnie, new PP_Variable[] {ann,a2}, p2);
            
            Potential p25 = new Potential(carl, new PP_Variable[] {bill,betsy}, p2);
            Potential p26 = new Potential(cecily, new PP_Variable[] {benny,bonnie}, p2);
            
            //Level 3:
            double[] p3 = {1, 0.5, 0.5, 0.25, 0, 0.5, 0.5, 0.5, 0, 0, 0, 0.25};
            Potential p31 = new Potential(dennis, new PP_Variable[] {carl,cecily}, p3);

            //Observation:
            Potential p41 = new Potential(dennis, new double[] {0, 0, 1});
            
            potentials = new Potential[] {p11, p12, p13, p14, p15, p21, p22, p23, p24, p25, p26, p31, p41};
        } 
        
        /*
         * Semiring implementation:
         */
        
        else if(args[1].equalsIgnoreCase("Semiring")) {
            
            // Level 1:
            double[] p1 = {0.99, 0.01};
            SRPotential p11 = new SRPotential(a1, p1);
            SRPotential p12 = new SRPotential(ann, p1);
            SRPotential p13 = new SRPotential(alan, p1);
            SRPotential p14 = new SRPotential(alice, p1);
            SRPotential p15 = new SRPotential(a2, p1);
            
            // Level 2:
            double[] p2 = {1, 0.5, 0.5, 0.33, 0, 0.5, 0.5, 0.67};
            SRPotential p21 = new SRPotential(bill, new PP_Variable[] {a1,ann}, p2);
            SRPotential p22 = new SRPotential(betsy, new PP_Variable[] {ann,alan}, p2);
            SRPotential p23 = new SRPotential(benny, new PP_Variable[] {alan,alice}, p2);
            SRPotential p24 = new SRPotential(bonnie, new PP_Variable[] {ann,a2}, p2);
            SRPotential p25 = new SRPotential(carl, new PP_Variable[] {bill,betsy}, p2);
            SRPotential p26 = new SRPotential(cecily, new PP_Variable[] {benny,bonnie}, p2);
            
            // Level 3:
            double[] p3 = {1, 0.5, 0.5, 0.25, 0, 0.5, 0.5, 0.5, 0, 0, 0, 0.25};
            SRPotential p31 = new SRPotential(dennis, new PP_Variable[] {carl,cecily}, p3);

            // Observation:
            SRPotential p41 = new SRPotential(dennis, new double[] {0, 0, 1});
            
            potentials = new SRPotential[] {p11, p12, p13, p14, p15, p21, p22, p23, p24, p25, p26, p31, p41};
        } 
        
        else {
            
            System.out.println("Wrong second Argument.");
            System.out.println("Usage: Studfarm int (Normal | Semiring) (Interactive | Console) (Local | Remote)");
            System.exit(1);
            
        }
        
        /*
         * Queries:
         */
        
        Domain[] queries = new Domain[] {
            new Domain(ann),
            new Domain(alan),
            new Domain(alice),
            new Domain(bill),
            new Domain(betsy),
            new Domain(benny),
            new Domain(bonnie),
            new Domain(carl),
            new Domain(cecily)
        };
                
        if(!args[3].equalsIgnoreCase("Local") && !args[3].equalsIgnoreCase("Remote")) {
            System.out.println("Wrong fourth Argument.");
            System.out.println("Usage: StudFarm int (Normal | Semiring) (Interactive | Console) (Local | Remote)");
            System.exit(1);
        }
                       
        /*
         * Setup NENOK Federation:
         */
              
        try {
                       
            // Run Processors :
            Processor[] processors = new Processor[procs]; 
           
            for (int i = 0; i < procs; i++) {
                processors[i] = (args[3].equalsIgnoreCase("Local")) ? Services.startLocalProcessor() : Services.startRemoteProcessor();
            }
            
            Locator[] locators = new Locator[potentials.length]; 
            
            // Store &amp; register potentials:
            for (int i = 0, j = 0; i < potentials.length; i++, j++) {
                locators[i] = processors[j % procs].store(potentials[i]);
            }
            
            Knowledgebase kb = null;
            if(args[3].equalsIgnoreCase("Remote")) {
            	
                // Create knowledgebase in Registry:
                Registry registry = Services.findRegistry();
                registry.createKnowledgebase(NAME);
                registry.addAll(NAME, locators);
                
                kb = registry.getKnowledgebase(NAME);
            	
            } else {
            	
            	// Create local knowledgebase object:
            	kb = new Knowledgebase(locators);
            	kb.setName(NAME);
            	
            }
            
            /*
             * Local Computation:
             */
            
            if(args[2].equalsIgnoreCase("Console")) {
            
                LCFactory factory = new LCFactory();
                factory.setArchitecture(Architecture.Lauritzen_Spiegelhalter);
                JoinTree tree = factory.create(kb, Arrays.asList(queries), processors[0]);
                tree.propagate();
                
                /*
                 * Query answering:
                 */
                
                System.out.println("Marginals of the StudFarm Example:\n");
                
                for(Domain query : queries) {
                    System.out.println(tree.answer(query));
                }
                            
            } else if(args[2].equalsIgnoreCase("Interactive")) {
                
            	if(args[3].equalsIgnoreCase("Remote"))
            		TreeViewer.display();
            	TreeViewer.display(kb);
                
            } else {
                System.out.println("Wrong third Argument.");
                System.out.println("Usage: StudFarm int (Normal | Semiring) (Interactive | Console) (Local | Remote)");
            }
                        
            /*
             * Destroy Processors:
             */
            
            if(args[3].equalsIgnoreCase("Remote")) {
            	Services.findRegistry().deleteKnowledgebase(NAME);
            }
            
            Services.destroyAllProcessors();

        } catch (Exception e1) {
            e1.printStackTrace();
            System.exit(1);
        }
    }
}
