package ppotentials;

import java.io.BufferedReader;
import java.io.File;
import java.io.FileNotFoundException;
import java.io.FileReader;
import java.io.IOException;
import java.io.StreamTokenizer;
import java.util.ArrayList;
import java.util.Hashtable;
import java.util.Iterator;

import nenok.parser.KB_Parser;
import nenok.parser.ParserException;
import nenok.parser.ResultSet;
import nenok.va.Domain;
import nenok.va.Valuation;

/**
 * A NENOK parser implementation that parses JavaBayes BIF files.
 * 
 * @author Marc Pouly
 * @version 1.0
 */

public class BIF_Parser implements KB_Parser {
	
	/**
	 * @see nenok.parser.Parser#accept(java.io.File)
	 */
	
	public boolean accept(File file) {
		return file.getName().endsWith(".bif");
	}

	/**
	 * @see nenok.parser.KB_Parser#parse(java.io.File)
	 */
	
	public ResultSet parse(File file) throws ParserException {
		try {
						
			BufferedReader in = new BufferedReader(new FileReader(file));
			StreamTokenizer tokenizer = new StreamTokenizer(in);
			
			tokenizer.resetSyntax();
			tokenizer.parseNumbers();
			tokenizer.wordChars('0', '9');
			tokenizer.wordChars('A', 'Z');
		    tokenizer.wordChars('a', 'z'); 
		    tokenizer.wordChars('_', '_'); 
		    tokenizer.wordChars('-', '-'); 
			tokenizer.slashSlashComments(true);
			
			/*
			 * Run tokenizer:
			 */
			
			int type;
			ArrayList<String> tokens = new ArrayList<String>();
		    while ((type = tokenizer.nextToken()) != StreamTokenizer.TT_EOF) {
		    	switch (type) {
					case StreamTokenizer.TT_WORD:
						 tokens.add(tokenizer.sval);
						 break;
					
					case StreamTokenizer.TT_NUMBER:
						 tokens.add(Double.toString(tokenizer.nval));
						 break;
				default:
					break;
				}
		    	
		    	if(tokenizer.ttype == ';' || tokenizer.ttype == ',' || 
		    	   tokenizer.ttype == '{' || tokenizer.ttype == '}' ||
		    	   tokenizer.ttype == '(' || tokenizer.ttype == ')' ||
		    	   tokenizer.ttype == '[' || tokenizer.ttype == ']') {
		    		tokens.add(new String(Character.toChars(tokenizer.ttype)));
		    		continue;
		    	}		    	
		    }
		    
			/*
			 * Parsing process:
			 */
		    
			Hashtable<String, PP_Variable> variables = new Hashtable<String, PP_Variable>();
			ArrayList<Potential> valuations = new ArrayList<Potential>();
			Iterator<String> iter = tokens.iterator();
			
			while(iter.hasNext()) {
				
				String token = iter.next();
				
				/*
				 * Parse Variables: 
				 */
				
				if(token.equalsIgnoreCase("Variable")) {
								
					String name = iter.next(); 
					
					String current;
					if(!(current = iter.next()).equals("{")) {
						throw new BIFException(current, "{");
					}
					if(!(current = iter.next()).equalsIgnoreCase("type")) {
						throw new BIFException(current, "type");
					}
					if(!(current = iter.next()).equalsIgnoreCase("discrete")) {
						throw new BIFException(current, "discrete");
					}
					
					if(!(current = iter.next()).equals("[")) {
						throw new BIFException(current, "[");
					}
								
					int count = (int)Double.parseDouble(iter.next());
						
					if(!(current = iter.next()).equals("]")) {
						throw new BIFException(current, "]");
					}		
					
					if(!(current = iter.next()).equals("{")) {
						throw new BIFException(current, "{");
					}
					
					String[] values = new String[count];
					for(int i = 0; i < count; i++) {
						values[i] = iter.next();
					}
					
					if(!(current = iter.next()).equals("}")) {
						throw new BIFException(current, "}");
					}
					
					while(!iter.next().equals("}")) {
						// do nothing ...
					}
					
					variables.put(name, new PP_Variable(name, values));
				}
				
				
				// Parse Valuation
				else if(token.equalsIgnoreCase("Probability")) {
										
					String current;
					if(!(current = iter.next()).equals("(")) {
						throw new BIFException(current, "(");
					}
					
					PP_Variable var = variables.get(iter.next());
					
					int frame = var.getFrame().length;
					
					ArrayList<PP_Variable> cond = new ArrayList<PP_Variable>();
					
					String str;
					while(!(str = iter.next()).equals(")")) {
						PP_Variable v = variables.get(str);
						if(v == null) {
							throw new BIFException("Unspecified variable "+v+".");
						}
						frame *= v.getFrame().length;
						cond.add(v);
					}
					
					if(!(current = iter.next()).equals("{")) {
						throw new BIFException(current, "{");
					}
					
					if(!(current = iter.next()).equalsIgnoreCase("table")) {
						throw new BIFException(current, "table");
					}
					
					double[] probs = new double[frame];
					current = iter.next();
					
					for(int i = 0; i < frame; i++) {
						probs[i] = Double.parseDouble(current);
						
						// Exponentials
						current = iter.next();
						if(current.startsWith("E")) {							
							int exp = Integer.parseInt(current.substring(2, current.length()));
							probs[i] = Math.pow(probs[i], exp);
							current = iter.next();
						} 
					}
										
					if(!current.equals(";")) {
						throw new BIFException(current, ";");
					}
					
					if(!(current = iter.next()).equals("}")) {
						throw new BIFException(current, "}");
					}
					
					if(cond.size() > 0) {
						valuations.add(new Potential(var, cond.toArray(new PP_Variable[cond.size()]), probs));
					} else {
						valuations.add(new Potential(var, probs));
					}
				}
			}
			
			in.close();
			return new ResultSet(valuations.toArray(new Valuation[valuations.size()]), new Domain[] {});
			
		} catch (FileNotFoundException e) {
			throw new ParserException("Parsing Error - File not found.", e);
		} catch (IOException e) {
			throw new ParserException("Parsing Error", e);
		}
	}

	/**
	 * Exception class for BIF parsing.
	 * 
	 * @author Marc Pouly
	 * @version 1.0
	 */
	
	public class BIFException extends ParserException {
		
		/**
		 * Constructor:
		 * @param message The error message.
		 */
		
		public BIFException(String message) {
			super(message);
		}
		
		/**
		 * Constructor:
		 * @param found The token that has been found.
		 * @param expected The expected token.
		 */
		
		public BIFException(String found, String expected) {
			super("Parsing Exception: Expected token is "+expected+", found "+found+".");
			System.out.println("Parsing Exception: Expected token is "+expected+", found "+found+".");
		}
		
	}
	
}
