package ppotentials;

import java.util.Hashtable;
import java.util.List;

import nenok.parser.ParserException;
import nenok.parser.XmlParser;
import nenok.va.Valuation;
import nenok.va.Variable;

import org.jdom.Element;

/**
 * This class provides a XML parser implementation for probability potentials.
 * 
 * @author Marc Pouly
 * @version 1.1
 */

public class PP_Parser extends XmlParser {

    /**
     * @see nenok.parser.XmlParser#parseVariable(java.lang.String, java.util.List)
     */
    
    public Variable parseVariable(String name, List<Element> content) throws ParserException {
		if(content == null || content.size() != 1) {
			throw new ParserException("Only <FRAME> tag allowed in <VARIABLE> tag.");
		}
		if(!(content.get(0)).getName().equals("frame")) {
			throw new ParserException("Wrong child in <VARIABLE>. <FRAME> must be specified.");
		}
		String[] frame = (content.get(0)).getText().split(" ");
    	return new PP_Variable(name, frame);
    } 
    
    /**
     * @see nenok.parser.XmlParser#parseValuation(java.util.List, java.util.Hashtable)
     */
    
    public Valuation parseValuation(List<Element> content, Hashtable<String, Variable> variables) throws ParserException {
    	if(content == null || content.size() != 2) {
    		throw new ParserException("Only <LABEL> and <PROBABILITIES> tags allowed in <VALUATION> tag.");
    	}
		if(!(content.get(0)).getName().equals("label")) {
			throw new ParserException("Wrong first child in <VALUATION> tag. <LABEL> must be specified.");
		}
    	PP_Variable[] vars = this.parseVariableString(content.get(0), variables);
		if(!(content.get(1)).getName().equals("probabilities")) {
			throw new ParserException("Wrong second child in <VALUATION> tag. <PROBABILITIES> must be specified.");
		}
        double[] probs = this.parseProbabilityString(content.get(1));
        return new Potential(vars, probs);
    }   
    
    /*
     * Helper methods:
     */
    
    /**
     * This method converts the information of the &quot;probabilities&quot; tag in the XML file
     * to an array of double values, i.e. probabilities.
     * @param element The content of the &quot;probabilities&quot; tag in the XML file.
     * @return A new probability array of double values.
     * @throws ParserException Exceptions caused by illegal number transformations.
     */
    
    private double[] parseProbabilityString(Element element) throws ParserException {
        String[] probs = element.getText().split(" ");
        double[] result = new double[probs.length];
        for(int i = 0; i < probs.length; i++) {
            try {
				result[i] = Double.parseDouble(probs[i]);
				if(result[i] < 0 || 1 < result[i]) {
					throw new ParserException("Probability value out of interval [0,1].");
				}
			} catch (NumberFormatException e) {
				throw new ParserException("Invalid probability value.");
			}            
        }
        return result;
    }
    
    /**
     * This method converts the information of the &quot;variables&quot; tag in the XML file
     * to an array of {@link PP_Variable} objects.
     * @param element The content of the &quot;variable&quot; tag in the XML file.
     * @param variables A hashtable that maps variable names on already constructed variable objects.
     * @return An array of variables that are contained in this potential.
     * @throws ParserException Exceptions caused by unknown variable specification.
     */    
    
    private PP_Variable[] parseVariableString(Element element, Hashtable<String, Variable> variables) throws ParserException {
        String[] vars = element.getText().split(" ");
        PP_Variable[] result = new PP_Variable[vars.length];
        for(int i = 0; i < vars.length; i++) {           
            result[i] = (PP_Variable)variables.get(vars[i]);
            if(result[i] == null) {
            	throw new ParserException("Unknown variable reference.");
            }
        }
        return result;
    }
}
