package ppotentials;

import java.util.Iterator;

import javax.swing.JComponent;

import nenok.va.Domain;
import nenok.va.Identity;
import nenok.va.Predictability;
import nenok.va.Predictor;
import nenok.va.Regularity;
import nenok.va.Scalability;
import nenok.va.VAException;
import nenok.va.Valuation;
import nenok.va.Variable;

/**
 * Probability potential are an example of a valuation algebra and this
 * class offers a possible implementation of the valuation interface for 
 * the case of probability potentials.
 * 
 * @author Marc Pouly
 * @version 1.1 
 */

public class Potential implements Predictability, Scalability {
        
    private PP_Variable[] variables;
    private double[] probabilities;
    private Domain domain;
        
    /*
     * Constructors:
     */
    
	/**
	 * Constructor:
	 * @param variable The main variable of this potential. 
	 * @param probabilities The probability for each possible configuration.<br/> 
	 * The number of probabilities must match with the product computed from each 
	 * variable's frame cardinality.
	 */
	
	public Potential(PP_Variable variable, double[] probabilities) {
	    if(variable == null || probabilities == null) {
	        throw new IllegalArgumentException("Assert arguments being != null.");
	    }
	    setInstanceFields(new PP_Variable[] {variable}, probabilities);
	}
	
	/**
	 * Constructor:
	 * @param variable The main variable of this potential. 
	 * @param conditionals The conditional variables of this potential.
	 * @param probabilities The probability for each possible configuration.<br/> 
	 * The number of probabilities must match with the product computed from each variable's frame cardinality.
	 */	
		
	public Potential(PP_Variable variable, PP_Variable[] conditionals, double[] probabilities) {
	    if(variable == null || conditionals == null || probabilities == null) {
	        throw new IllegalArgumentException("Assert arguments being != null.");
	    }
	    PP_Variable[] all = new PP_Variable[conditionals.length+1];
	    all[0] = variable;
		for (int i = 1; i < all.length; i++) {
			all[i] = conditionals[i-1];
		}
	    setInstanceFields(all,probabilities);
	}
	
	/**
	 * Private constructor:
	 * Sets instance fields and performs dimension check.
	 * @param variables An array of discrete variables.
	 * @param probabilities An array of probabilities.
	 */
	
	protected Potential(PP_Variable[] variables, double[] probabilities) {
	    if(variables == null || probabilities == null) {
	        throw new IllegalArgumentException("Assert arguments being != null.");
	    }
	    setInstanceFields(variables, probabilities);
	}
	
	/**
	 * Helper method shared by multiple constructors.
	 * Sets instance fields and performs dimension check.
	 * @param variables An array of discrete variables.
	 * @param probabilities An array of probabilities.
	 */
 
	private void setInstanceFields(PP_Variable[] variables, double[] probabilities) {
	    this.variables = variables;
	    this.probabilities = probabilities;
	    if(countConfigurations() != probabilities.length) {
	        throw new IllegalArgumentException("Dimension of probability array doesn't match. Given: "+probabilities.length+", expected: "+countConfigurations()+".");
	    }
	}

    /*
     * Valuation implementation:
     */
    
    /**
     * @see nenok.va.Valuation#label()
     */
	
    @SuppressWarnings("all")
    public Domain label() {
    	if(domain == null) {
    		domain = new Domain(variables);
    	}
        return domain;
    }

    /**
     * @see nenok.va.Valuation#combine(nenok.va.Valuation)
     */
    
    public Valuation combine(Valuation val) {
		if(val instanceof Identity) {
			return this;
		}
        Potential arg = (Potential)val;
        //Number of variable in the new domain:
        int inter = 0;
        for (int i = 0; i < variables.length; i++) {
			if(indexOf(variables[i], arg.variables) >= 0) {
				inter++;
			}
		}
		//Variable array of the new potential:
		PP_Variable[] newVars = new PP_Variable[arg.variables.length + this.variables.length - inter];
		Variable[] interVars = new Variable[inter];
        
		//Fill new variable array and compute dimension of probability array:
		int probabilitySize = 1, newVarsCounter = 0, interCounter = 0;
		for (; newVarsCounter < this.variables.length; newVarsCounter++) {
			newVars[newVarsCounter] = this.variables[newVarsCounter];
			probabilitySize *= this.variables[newVarsCounter].getFrame().length;
		}
		for (int j = 0; j < arg.variables.length; j++) {
			if(indexOf(arg.variables[j], variables) < 0) {
				newVars[newVarsCounter++] = arg.variables[j];
				probabilitySize *= arg.variables[j].getFrame().length;
			} else {
				interVars[interCounter++]= arg.variables[j];
			}
		}
		
		//Probability array of the new potential:
		double[] newProbabilities = new double[probabilitySize];
		
		/*
		 * Compute each configuration's value
		 */
			
		int[] symbols = new int[inter];
		int thisConfgurations = countConfigurations();
		int argConfigurations = arg.countConfigurations();
		
		// Determine indexes of intervars:
		int thisIndex[] = new int[symbols.length];
		int argIndex[] = new int[symbols.length];
		for (int i = 0; i < interVars.length; i++) {
			thisIndex[i] = indexOf(interVars[i], variables);
			argIndex[i] = indexOf(interVars[i], arg.variables);
		}
		
		for (int i = 0, configurationIndex = 0; i < thisConfgurations; i++) {
		    //Store all variables' indexes of this.variables:
		    for (int j = 0; j < interVars.length; j++) {
		    	symbols[j] = getSymbolIndex(thisIndex[j], i);
            }
		    for (int j = 0; j < argConfigurations; j++) {
		        boolean multiply = true;
		        for (int k = 0; k < interVars.length; k++) {
                    multiply &= (arg.getSymbolIndex(argIndex[k], j) == symbols[k]);
                }
		        //Perform multiplication:
		        if(multiply) {
		            newProbabilities[configurationIndex++] = this.probabilities[i]*arg.probabilities[j];
		        }
            }
        }
		return new Potential(newVars, newProbabilities);
    }

    /**
     * @see nenok.va.Valuation#marginalize(nenok.va.Domain)
     */
    
	public Valuation marginalize(final Domain dom) {
		if(label().equals(dom)) {
		    return this;
		}
		if(!dom.subSetOf(label())) {
			new VAException("Marginalization impossible: this.label() = "+label()+", argument = "+dom+".");
		}
		Potential result = this;
		//Calculates the difference set between both domains.
		Domain difference = Domain.difference(label(), dom);
		Iterator it = difference.iterator();
		//Remove each variable in the difference set.
		while(it.hasNext()) {
			PP_Variable var = (PP_Variable)it.next();
			result = result.eliminateVariable(var);
		}
		return result;
	}
	    
    /**
     * @see nenok.va.Valuation#weight()
     */
    
    public int weight() {
        return predictor().predict(this.label());
    }
    
    /*
     * Predictability implementation:
     */
    
    /**
     * @see nenok.va.Predictability#predictor()
     */ 
    
    public Predictor predictor() {
        return PP_Predictor.getInstance();
    }
        
    /*
     * Dividability implementation:
     */
        
    /**
     * @see nenok.va.Regularity#inverse()
     */
    
    public Regularity inverse() {
        double probs[] = new double[this.probabilities.length];
        for (int i = 0; i < probs.length; i++) {
            if(!Probability.numericTest(this.probabilities[i], 0.0)) {
                probs[i] = 1 / this.probabilities[i];
            } else {
            	probs[i] = 0.0;
            }
        }
        return new Potential(this.variables, probs);
    }
          
    /*
     * Scalability implementation:
     */
    
    /**
     * @see nenok.va.Scalability#scale()
     */
    
    public Scalability scale() {
        return new Scalability.Implementor().scale(this);
    }
    
	/*
	 * Equality checks:
	 */
	
    /**
     * @see java.lang.Object#equals(java.lang.Object)
     */
	
	public boolean equals(Object o) {
		if(!(o instanceof Potential)) {
			return false;
		}
		if(o == this) {
			return true;
		}
		Potential p = (Potential)o;
		if(variables.length != p.variables.length || probabilities.length != p.probabilities.length) {
			return false;
		}		
		//Slow test:
		int[] indexes = new int[variables.length];
		for (int i = 0; i < probabilities.length; i++) {
		    //Fill indexes array:
		    for (int j = 0; j < indexes.length; j++) {
                indexes[j] = getSymbolIndex(j,i);
            }
		    //Permute indexes array:
		    int[] permutation = new int[variables.length];	    
		    for (int j = 0; j < variables.length; j++) {
                for (int k = 0; k < p.variables.length; k++) {
                    if(variables[j].equals(p.variables[k])) {
                        permutation[k] = indexes[j];
                    }
                }
            }
		    //Compare with argument:
		    int argConfigurations = p.countConfigurations();
		    for (int j = 0; j < argConfigurations; j++) {
		        boolean correct = true;
		        for (int k = 0; k < permutation.length; k++) {
		            int index = indexOf(p.variables[k], p.variables);
                    correct &= (p.getSymbolIndex(index, j) == permutation[k]);
                }
		        //Perform multiplication:
		        if(correct) {
					if(!Probability.numericTest(this.probabilities[i], p.probabilities[j])) {
					    return false;
					}
					break;
				}
            }
        }
		return true;
	}
			
    /*
     * Output methods:
     */
    
    /**
     * @return The current potential as a string.
     */
    
    public String displayText() {
        int max = 0;
        String format = "", result = "", line = "", title = "Prob.";
        
        /*
         * Find string of maximum length:
         */
        
        for (int i = 0; i < variables.length; i++) {
            PP_Variable v = variables[i];
            if(max < v.toString().length()) {
                max = v.toString().length();
            }
            for (int j = 0; j < v.getFrame().length; j++) {
                if(max < v.getFrame()[j].toString().length()) {
                    max = v.getFrame()[j].toString().length();
                }
            }
        }
        
        if(max < title.length()) {
            max = title.length();
        }
        
        max += 2;
        
        /*
         * Prepare format string for title:
         */

        for (int i = 0; i < variables.length+1; i++) {
        	format += "|%"+(i+1)+"$-"+max+"s";
        }
        format += "|\n";
        
        /*
         * Output header line:
         */
        
        Object[] data = new Object[variables.length+1];
        for (int i = 0; i < variables.length; i++) {
            data[i] = variables[i];
        }
        data[variables.length] = title;
        result += String.format(format, data);
        
        /*
         * Horizontal line:
         */
        
        for (int i = 0; i < result.length()-1; i++) {
            line += "-";
        }
        line += "\n";
        result = line+result+line;

        /*
         * Prepare format string for configurations:
         */

        format = "";
        for (int i = 0; i < variables.length; i++) {
        	format += "|%"+(i+1)+"$-"+max+"s";
        }
        format += "|%"+(variables.length+1)+"$-"+max+".3f|\n";
                
        /*
         * Configurations:
         */
        
        for (int i = 0; i < probabilities.length; i++) {
            Object row[] = new Object[data.length];
            for (int j = 0; j < variables.length; j++) {
                row[j] = variables[j].getFrame()[getSymbolIndex(j,i)].toString();
            }
            row[data.length-1] = probabilities[i];
            result += String.format(format, row);
        }   
        
        result += line;
        
		return result;
    }
        
    /**
     * Displays the current potential as a chart.
     * @return The current potential as a chart.
     */
    
    public JComponent displayChart() {
        return new Chart(this);
    }
    
    /**
     * @see java.lang.Object#toString()
     */
    
    public String toString() {
        return displayText();
    }
    	    
    /*
     * Helper methods:
     */
    
    /**
     * Removes the given variable from the current potential i.e. calculates the marginalization
     * onto <code>dom = this.dom - {var}</code>. Generates a new potential instance - the original
     * potential stays untouched.
     * @param var The variable to remove.
     * @return The resulting potential.
     */
    
    private Potential eliminateVariable(PP_Variable var) {
        //The new variable array.
        PP_Variable[] newVars = new PP_Variable[this.variables.length - 1];
        int oldPosition = -1;
        int size = 1;
        //Fill the new variable array and calculate the number of configurations of the new potential.
        for (int i = 0, j = 0; i < this.variables.length; i++) {
            if(!this.variables[i].equals(var)) {
                size *= this.variables[i].getFrame().length;
                newVars[j++] = this.variables[i];
            } else {
                oldPosition = i;
            }
        }
        //The new probability array.
        double[] newProbabilities = new double[size];
        int offset = 1;
        for (int i = oldPosition + 1; i < variables.length; i++) {
            offset *= this.variables[i].getFrame().length;
        }
        //Calculate the elements of the new probability array.
        int blocs = (probabilities.length / var.getFrame().length) / offset;
        int blocLines = newProbabilities.length / blocs;
        for (int i = 0, index = 0; i < blocs; i++) {
            int base = i * (probabilities.length / blocs);
            for (int j = 0; j < blocLines; j++) {
                double sum = 0.0;               
                for (int k = 0; k < var.getFrame().length; k++) {   
                    sum += this.probabilities[base + j + k*offset];
                }
                newProbabilities[index] = sum;
                index++;
            }
        }
        //Generate and return the new instance.
        return new Potential(newVars, newProbabilities);
    }
    
    /**
     * This method returns the index of the variable's value at a specifique configuration.
     * @param varPos The variable caracterized by its position.
     * @param configuration The configuration
     * @return Ths symbol of the variable at the given configuration.
     * Example:<br/>
     * <code>getSymbol(0,3)</code> returns the index of variable 0's symbol at configuration 3.
     */
    
    protected int getSymbolIndex(int varPos, int configuration) {
        int repetitions = 1;
        for (int i = varPos + 1; i < variables.length; i++) {
            repetitions *= variables[i].getFrame().length;
        }
        return (configuration / repetitions) % variables[varPos].getFrame().length;
    }
        
    /**
     * @return Compute the number of configurations.
     */
    
    private int countConfigurations() {
        int size = 1;
        for (int i = 0; i < variables.length; i++) {
            size *= variables[i].getFrame().length;
        }
        return size;        
        
    }
    
    /**
     * Returns the index of the given variable in the variable array.
     * @param variable The variable whose index is searched.
     * @param variables The variable array which contains the index.
     * @return the variable's index or <code>-1</code>.
     */
    
    private static int indexOf(Variable variable, Variable[] variables) {
    	for (int i = 0; i < variables.length; i++) {
            if(variables[i].equals(variable)) {
                return i;
            }
        }
        return -1;
    }
    
    /*
     * Getter:
     */

    /**
     * @return Returns the variables of this potential.
     */

    protected PP_Variable[] getVariables() {
        return variables;
    }

    /**
     * @return Returns the probability array of this potential. 
     */

    public double[] getProbabilities() {
        return probabilities;
    }
}
