package nenok;

import java.rmi.RemoteException;
import java.util.Collection;

import nenok.lc.JoinTree;
import nenok.lc.constr.Algorithm;
import nenok.lc.constr.ConstrException;
import nenok.lc.hugin.HJoinTree;
import nenok.lc.id.IDJoinTree;
import nenok.lc.ls.LSJoinTree;
import nenok.lc.ss.BSSJoinTree;
import nenok.lc.ss.SSJoinTree;
import nenok.net.Knowledgebase;
import nenok.net.processor.Processor;
import nenok.va.Domain;

/**
 * This class serves as a more comfortable user interface
 * to choose local computation architectures.
 * 
 * @author Marc Pouly
 * @version 1.1
 */

public enum Architecture {
    
    /**
     * The n-ary Shenoy-Shafer architecture.
     */
    
    Shenoy_Shafer,
    
    /**
     * The binary Shenoy-Shafer architecture.
     */
    
    Binary_Shenoy_Shafer,
    
    /**
     * The Lauritzen-Spiegelhalter architecture.
     */
    
    Lauritzen_Spiegelhalter,
    
    /**
     * The Hugin architecture.
     */
    
    Hugin,
    
    /**
     * The idempotent architecture.
     */
    
    Idempotent_Architecture;
    
    /**
     * This method maps enum types on new jointree instances.
     * @param kb The knoelwedgebase for the new jointree.
     * @param queries The query list for the new jointree.
     * @param algo The construction algorithm to build the jointree.
     * @param proc The processor assigned to query nodes.
     * @return A new jointree instance.
     * @throws RemoteException Exceptions caused by communication problems when the query processor is accessed.
     * @throws ConstrException Exceptions thrown by the jointree construction process.
     */
    
    public JoinTree getInstance(Knowledgebase kb, Collection<Domain> queries, Algorithm algo, Processor proc) throws RemoteException, ConstrException {
        switch(this) {
            case Binary_Shenoy_Shafer: return new BSSJoinTree(kb, queries, algo, proc); 
            case Shenoy_Shafer: return new SSJoinTree(kb, queries, algo, proc); 
            case Lauritzen_Spiegelhalter: return new LSJoinTree(kb, queries, algo, proc); 
            case Hugin: return new HJoinTree(kb, queries, algo, proc); 
            case Idempotent_Architecture: return new IDJoinTree(kb, queries, algo, proc); 
            default: return null;
        }
    }
    
    /**
     * @return The name of the architecture represented by this enum type.
     */
    
    public String getName() {
        switch(this) {
            case Binary_Shenoy_Shafer: return "Binary Shenoy-Shafer"; 
            case Shenoy_Shafer: return "Shenoy-Shafer"; 
            case Lauritzen_Spiegelhalter: return "Lauritzen-Spiegelhalter"; 
            case Hugin: return "Hugin"; 
            case Idempotent_Architecture: return "Idempotent Architecture"; 
            default: return "Unknown Architecture";
        }
    } 
}
