package nenok;

import java.io.File;
import java.rmi.RemoteException;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collection;

import nenok.lc.JoinTree;
import nenok.lc.constr.Algorithm;
import nenok.lc.constr.ConstrException;
import nenok.lc.constr.Deserializer;
import nenok.lc.constr.OSLA_SC;
import nenok.net.Knowledgebase;
import nenok.net.processor.Processor;
import nenok.parser.JT_Parser;
import nenok.parser.KB_Parser;
import nenok.parser.ParserException;
import nenok.parser.ResultSet;
import nenok.va.Domain;
import nenok.va.Valuation;

/**
 * FACTORY class to build jointrees. Various methods to build join trees with different parameter lists are offered 
 * by this class. Basically, its main purposes is to simplify the user's life by preventing him from the details of 
 * using default values as parameters.
 *  
 * @author Marc Pouly
 * @version 1.2
 */

public final class LCFactory {
    
    private Algorithm algo;
    private Architecture architecture;
    
    /*
     * Constructors:
     */
    
    /**
     * Constructor:
     */
    
    public LCFactory() {
    	this(Architecture.Binary_Shenoy_Shafer);
    }
            
    /**
     * Constructor:
     * Constructs jointrees of the given architecture class with an OSLA_SC construction algorithm. 
     * @param architecture The architecture of local computation, i.e. the corresponding architecture's enum type.
     */
    
    public LCFactory(Architecture architecture) {
        this.algo = new OSLA_SC();
        this.architecture = (architecture == null)? Architecture.Binary_Shenoy_Shafer : architecture;
    }
    
    /*
     * Create methods with Knowledgebases:
     */
    
    /**
     * Create method of this FACTORY class. Builds a jointree that corresponds to this factory's architecture from the given 
     * knowledgebase and query list. The factory's construction algorithm is used to build the tree. 
     * @param kb The knowledgebase from whom the jointree is build.
     * @param queries A list of domains that are ensured to be covered by the resulting jointree.
     * @param proc The processor that shall be assigned to query nodes.
     * @return A jointree instance that corresponds to this factory's architecture.
     * @throws RemoteException Exceptions caused by communication problems when the factory's processor is accessed.
     * @throws ConstrException Exceptions thrown by the jointree construction process.
     */
    
    public JoinTree create(Knowledgebase kb, Collection<Domain> queries, Processor proc) throws RemoteException, ConstrException {
        return architecture.getInstance(kb, queries, algo, proc);
    }
    
    /**
     * Create method of this FACTORY class. Builds a jointree that corresponds to this factory's architecture from the given 
     * knowledgebase and query list. The factory's construction algorithm is used to build the tree. 
     * @param kb The knowledgebase from whom the jointree is build.
     * @param queries An array of domains that are ensured to be covered by the resulting jointree.
     * @param proc The processor that shall be assigned to query nodes.
     * @return A jointree instance that corresponds to this factory's architecture.
     * @throws RemoteException Exceptions caused by communication problems when the factory's processor is accessed.
     * @throws ConstrException Exceptions thrown by the jointree construction process.
     */
    
    public JoinTree create(Knowledgebase kb, Domain[] queries, Processor proc) throws RemoteException, ConstrException {
        return architecture.getInstance(kb, Arrays.asList(queries), algo, proc);
    }
    
    /**
     * Create method of this FACTORY class. Builds a jointree that corresponds to this factory's architecture from the given 
     * knowledgebase and query list. The factory's construction algorithm is used to build the tree. An arbitrary processor
     * that occurs in the knowledgebase is assigned to query nodes.
     * @param kb The knowledgebase from whom the jointree is build.
     * @return A jointree instance that corresponds to this factory's architecture.
     * @throws RemoteException Exceptions caused by communication problems when the factory's processor is accessed.
     * @throws ConstrException Exceptions thrown by the jointree construction process.
     */
    
    public JoinTree create(Knowledgebase kb) throws RemoteException, ConstrException {
        return architecture.getInstance(kb, new ArrayList<Domain>(), algo, null);
    }

    /*
     * Create methods with Files:
     */
    
    /**
     * Create method of this FACTORY class. Builds a jointree that corresponds to this factory's architecture
     * from the parsed file content. The parsed {@link Valuation} objects will be stored on the processor given
     * as argument. The factory's construction algorithm is used to build the tree.
     * @param file The file to parse.
     * @param parser The user defined parser instance.
     * @param proc The processor that shall be assigned to query nodes.
     * @return A jointree instance that corresponds to this factory's architecture.
     * @throws RemoteException Exceptions caused by communication problems when the factory's processor is accessed.
     * @throws ConstrException Exceptions thrown by the jointree construction process.
     * @throws ParserException Exceptions caused by the parsing process.
     */
    
    public JoinTree create(File file, KB_Parser parser, Processor proc) throws RemoteException, ConstrException, ParserException {
        if(!parser.accept(file)) {
        	throw new ParserException("This parser does not accept the given file.");
        }
    	ResultSet result = parser.parse(file);
        Knowledgebase kb = new Knowledgebase(proc.store(result.vals));
        return create(kb, Arrays.asList(result.queries), proc);
    }
        
    /**
     * Deserializes a jointree from a file by use of the given parser instance.
     * @param file The file containing a serialized jointree.
     * @param parser The parser for the deserialization process.
     * @param proc A processor instance to upload the parsed valuations.
     * @return A jointree instance that corresponds to this factory's architecture.
     * @throws RemoteException Exceptions caused by communication problems when the factory's processor is accessed.
     * @throws ConstrException Exceptions thrown by the jointree construction process.
     */
   
    public JoinTree rebuild(File file, JT_Parser parser, Processor proc) throws RemoteException, ConstrException {
        if(!parser.accept(file)) {
        	throw new ConstrException("This parser does not accept the given file.");
        }
    	Algorithm tmp = this.algo;
        this.algo = new Deserializer(file, parser);
        JoinTree tree = create(new Knowledgebase(), new ArrayList<Domain>(), proc);
        this.algo = tmp;
        return tree;
    }

    /*
     * Getter &amp; Setter:
     */
    
    /**
     * Method to change the current factory's local computation architecture.
     * @param architecture The new local computation architecture.
     */
    
    public void setArchitecture(Architecture architecture) {
        this.architecture = architecture;
    }

    /**
     * Method to change the current factory's jointree construction algorithm.
     * @param algo The new jointree construction algorithm.
     */
    
    public void setConstructionAlgorithm(Algorithm algo) {
        this.algo = algo;
    }
}
