package nenok;

import java.net.InetAddress;
import java.net.MalformedURLException;
import java.net.UnknownHostException;
import java.rmi.RMISecurityManager;
import java.rmi.RemoteException;
import java.util.Hashtable;
import java.util.logging.Level;
import java.util.logging.Logger;

import nenok.net.Locator;
import nenok.net.processor.Processor;
import nenok.net.processor.ProcessorImpl;
import nenok.net.processor.Starter;
import nenok.net.registry.Registry;
import net.jini.core.discovery.LookupLocator;
import net.jini.core.entry.Entry;
import net.jini.core.lookup.ServiceItem;
import net.jini.core.lookup.ServiceTemplate;
import net.jini.discovery.DiscoveryGroupManagement;
import net.jini.discovery.LookupDiscoveryManager;
import net.jini.discovery.LookupLocatorDiscovery;
import net.jini.id.Uuid;
import net.jini.id.UuidFactory;
import net.jini.lease.LeaseRenewalManager;
import net.jini.lookup.LookupCache;
import net.jini.lookup.ServiceDiscoveryManager;
import net.jini.lookup.entry.Name;

/**
 * This utility class deals with Nenok services. It provides static methods to start, localize &amp; terminate
 * Nenok {@link Processor}s. Furthermore, localization of {@link Registry} services is an important task of 
 * this class, which is implemented as multicast localization. To assert best possible lookup times a static 
 * JINI {@link LookupCache} is used.
 * 
 * @author Marc Pouly
 * @version 1.2
 */

public class Services {
        
    private static LookupCache cache;
    private static LookupDiscoveryManager mgr;
    private static ServiceDiscoveryManager sdm;
    
    private static Hashtable<Uuid, Processor> localProcs;
    private static Hashtable<Uuid, Starter> remoteProcs;
    
    /*
     * Constants:
     */
    
    private static final int TIME = 5000;
    private static final String[] GROUP = DiscoveryGroupManagement.ALL_GROUPS;
    
    /*
     * Templates:
     */
    
	private static final ServiceTemplate REG_TMP = new ServiceTemplate(null, new Class[] {Registry.class}, null);
    
    /*
     * Ensure that a SecurityManager is running.
     */
    
    static {
        if(System.getSecurityManager() == null) {
            System.setSecurityManager(new RMISecurityManager());
        }
        
        try {
        	System.getProperty("java.policy");
        } catch(Exception e) {
        	System.err.println("Security problem: No policy file set.\n");
        	System.exit(1);
        }
        
        /*
         * Turn off JINI internal logger.
         */
        
        Logger.getLogger(LookupLocatorDiscovery.class.getName()).setLevel(Level.OFF);

        remoteProcs = new Hashtable<Uuid, Starter>();
        localProcs = new Hashtable<Uuid, Processor>();
    }
    
    /**
     * Jini environment setup &amp; service caching.
     */
    
    private static void init() {
    	if(mgr != null && cache != null) {
    		return;
    	}
        try {
            mgr = new LookupDiscoveryManager(GROUP, null, null);
            sdm = new ServiceDiscoveryManager(mgr, new LeaseRenewalManager());
            cache = sdm.createLookupCache(null, null, null);
        } catch (Exception e) {
            e.printStackTrace();
            System.exit(1);
        } 
    }

	/*
	 * Service localization methods:
	 */
	 
    /**
     * Finds the processor that hosts the valuation referenced by the given locator.
     * @param locator The locator whose processor is to localize.
     * @return The referenced processor or <code>null</code>, if locatozation failed.
     */
    
    public static Processor findProcessor(Locator locator) {  	
        return findProcessor(locator.getPID());
    }
    
    /**
     * Finds the processor with the given ID, running of the given host.
     * @param pid The processor identifier.
     * @return The referenced processor or <code>null</code>, if locatozation failed.
     */
    
    public static Processor findProcessor(Uuid pid) {
    	
    	/*
    	 * For local processors:
    	 */
    	
    	if(localProcs.containsKey(pid)) {
    		return localProcs.get(pid);
    	}
    	
    	/*
    	 * For remote processors:
    	 */
    	
    	init();
                        
        /*
         * Service localization:
         */
    	
    	Entry[] attributes = new Entry[] {new Name(pid.toString())};
        ServiceTemplate tmp = new ServiceTemplate(null, new Class[] {Processor.class}, attributes);
    	
        try {
        	ServiceItem item = sdm.lookup(tmp, null, 5000);

        	if (item != null) {
        	  return (Processor)item.service;
        	}
        	
    	} catch(Exception e) {
    	    e.printStackTrace();
    	} 
    	
    	return null;
    }

    /**
     * Starts a new remote processor instance.
     * @return The proxy of the created processor.
     * @throws UnknownHostException Exception thrown when the machine's host name cannot be localized.
     * @deprecated Use {@link Services#startRemoteProcessor} instead.
     */
    
    public static Processor startProcessor() throws UnknownHostException {
    	return startRemoteProcessor();
    }
    
    /**
     * Starts a new remote processor instance.
     * @return The proxy of the created processor.
     * @throws UnknownHostException Exception thrown when the machine's host name cannot be localized.
     */
    
    public static Processor startRemoteProcessor() throws UnknownHostException {
    	Uuid pid = UuidFactory.generate();
        Starter starter = new Starter(pid);
        starter.start();
        remoteProcs.put(pid, starter);
        return findProcessor(pid);
    }
    
    /**
     * Starts a new local processor thread.
     * @return The reference to the created processor running in a new thread.
     */
     
    public static Processor startLocalProcessor() {
        Uuid pid = UuidFactory.generate();  
        Processor proc = new ProcessorImpl(pid);
        localProcs.put(pid, proc);
        return proc;
    }
    
    /**
     * Destroys a given processor. This method unexports the processor service
     * and cancels all leases. The service disappears when the leases expire.
     * @param pid The identifier of the processor to destroy.
     * @return <code>true</code>, if the processor has been destroyed successfully.
     */
    
    public static boolean destroyProcessor(Uuid pid) {
    	
    	if(pid == null) {
    		return false;
    	}
    	
    	/*
    	 * For local processors:
    	 */
    	
    	if(localProcs.containsKey(pid)) {
    		localProcs.remove(pid);
    		return true;
    	}
    	
    	/*
    	 * For remote processors:
    	 */
    	
        Starter starter = remoteProcs.remove(pid);
        return (starter == null)? false : starter.terminate();
    }
    
    /**
     * Destroys a given processor. This method unexports the processor service
     * and cancels all leases. The service disappears when the leases expire.
     * @param proc The processor to destroy.
     * @return <code>true</code>, if the processor has been destroyed successfully.
     */
    
    public static boolean destroyProcessor(Processor proc) {
        try {
            return (proc != null) && destroyProcessor(proc.getPID());
        } catch (RemoteException e) {
            return false;
        }
    }
    
    /**
     * Destroy all processors that were started from this class.
     * @return <code>true</code>, if all processors have been destroyed successfully.
     */
    
    public static boolean destroyAllProcessors() {
        boolean result = true;
        for(Starter starter : remoteProcs.values()) {
            result &= starter.terminate();
        }
        remoteProcs.clear();
        localProcs.clear();
        return result;
    }
    
    /**
     * Finds an arbitrary registry service via multicast.
     * @return A registry service or <code>null</code>, if localization failed.
     */
    
    public static Registry findRegistry() {
	       
    	init();
    	     	
    	/*
    	 * Multicast localization:
    	 */
    	
    	mgr.setLocators(new LookupLocator[0]);
    	    	
        /*
         * Service localization:
         */
    	
    	try {
    	
    		ServiceItem item = sdm.lookup(REG_TMP, null, TIME);

    	    if (item != null) {
    	    	return (Registry)item.service;
    	    }
    	    
    	} catch(Exception e) {
    	    // do nothing
    	} 
    	
    	return null;
    }  
    
    /**
     * Finds an arbitrary registry service via unicast. 
     * Thereby, only registry services running on the given URLs are returned.
     * @param urls URLs for unicast localization beyond subnets.
     * @return A registry service or <code>null</code>, if localization failed.
     * @throws MalformedURLException Invalide URL argument.
     * @throws UnknownHostException URL not resolvable.
     */
        
    public static Registry findRegistry(String... urls) throws MalformedURLException, UnknownHostException {
    	       
    	if(urls.length == 0) {
    		throw new IllegalArgumentException("Specify at least one URL for unicast locating.");
    	}
    	
    	init();
    	
    	/*
    	 * URL verification:
    	 */
    	
    	for(String url : urls) {
    		if(!url.startsWith("jini://")) {
    			throw new MalformedURLException("Malformed URL: Use JINI protocol.");
    		}
    		
    		InetAddress address = InetAddress.getByName(url.substring("jini://".length()));
    		if(address == null) {
    			throw new UnknownHostException("Unknown Host: Cannot locate "+url+".");
    		}
    	}
    	
    	/*
    	 * Unicast localization:
    	 */

    	LookupLocator[] locs = new LookupLocator[urls.length];
    	for (int i = 0; i < locs.length; i++) {
    		locs[i] = new LookupLocator(urls[i]);
		}
    	mgr.setLocators(locs);
    	
        /*
         * Service localization:
         */
    	
    	try {
    	
    		ServiceItem item = sdm.lookup(REG_TMP, null, TIME);

    	    if (item != null) {
    	    	
    	    	Registry registry = (Registry)item.service;
    	    	InetAddress regHost = InetAddress.getByName(registry.getIP());
    	    	for(LookupLocator loc : mgr.getLocators()) {
    	    		if(regHost.equals(InetAddress.getByName(loc.getHost()))) {
    	    			return registry;
    	    		}
    	    	}
    	    	
    	    	return (Registry)item.service;
    	    }
    	    
    	} catch(RemoteException e) {
    	    // do nothing ...
    	} catch(InterruptedException e) {
    	    // do nothing ...
    	}
    	
    	return null;
    }  
}