package nenok;

import java.lang.reflect.Method;
import java.util.ArrayList;
import java.util.List;

import javax.swing.JComponent;

import nenok.va.Valuation;

/**
 * Utility class that provides a collection of static methods which extend the java reflection framework.
 * 
 * @author Marc Pouly
 * @version 1.2
 */

public class Utilities {
	
    /**
     * Helper method to determine if a given class implements a given interface.
     * @param _interface Is this interface implemented by the given class.
     * @param _class Does this class implement the given interface.
     * @return <code>true</code>, if the given class implements the given interface.
     */
    
    public static boolean isInterfaceOf(Class _interface, Class _class) {
        Class c = _class;
        while(c != null) {
            if(findInterface(_interface, c.getInterfaces())) {
                return true;
            }
            c = c.getSuperclass();
        }
        return false;
    }
    
    /**
     * Helper method to determine if a given class is superclass of another.
     * Convention: Every class is superclass of itself.
     * @param _super Is this class ascendent of the second argument.
     * @param _class Is this class descendent of the former argument.
     * @return <code>true</code>, if the first argument is a superclass of the second.
     */
    
    public static boolean isSuperclassOf(Class _super, Class _class) {
        Class superclass = _class;
    	while (superclass != null) {
            if(superclass.getName().equals(_super.getName())) {
            	return true;
            }
            superclass = superclass.getSuperclass();
         }
    	return false;
    }
    
    /**
     * This static method returns a list of all string representation functions defined in the
     * argument's class. A string representation method's name starts with &quot;display&quot;,
     * returns a string and has an empty no parameter list.
     * @param val The valuation class instance to be searched.
     * @return All string representation methods contained in the argument's class.
     */
    
    public static List<Method> getStringRepresentators(Valuation val) {
        Class c = val.getClass();
        ArrayList<Method> result = new ArrayList<Method>();
        for(Method m : c.getMethods()) {
            if(m.getName().startsWith("display") && m.getParameterTypes().length == 0 && m.getReturnType().equals(String.class)) {
                    result.add(m);
            }
        }
        return result;
    }
    
    /**
     * This static method returns a list of all graphical representation functions defined in the
     * argument's class. A graphical representation method's name starts with &quot;display&quot;,
     * returns a SWT composite and has the parent composite as only parameter.
     * @param val The valuation class instance to be searched.
     * @return All graphical representation methods contained in the argument's class.
     */
    
    public static List<Method> getGraphicalRepresentators(Valuation val) {
        Class c = val.getClass();
        ArrayList<Method> result = new ArrayList<Method>();
        for(Method m : c.getMethods()) {
            if(m.getName().startsWith("display") && isSuperclassOf(JComponent.class, m.getReturnType())) {
                result.add(m);
            }
        }
        return result;
    }
    
    /**
     * Is an interface contained in an array of interfaces or in their super interfaces.
     * @param theInterface The interface to find.
     * @param interfaces An array of interfaces.
     * @return <code>true</code>, if the interface has been found.
     */
    
    private static boolean findInterface(Class theInterface, Class[] interfaces) {
        for(Class i : interfaces) {
            if(i.equals(theInterface)) {
                return true;
            }
            if(findInterface(theInterface, i.getInterfaces())) {
                return true;
            }
        }
        return false;
    }
}
