package nenok.lc;

import java.rmi.RemoteException;
import java.util.ArrayList;
import java.util.Collection;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Set;

import nenok.Services;
import nenok.Utilities;
import nenok.lc.alloc.Allocator;
import nenok.lc.alloc.Default;
import nenok.lc.alloc.Multiway;
import nenok.lc.constr.Algorithm;
import nenok.lc.constr.ConstrException;
import nenok.lc.constr.OSLA_SC;
import nenok.lc.out.AsciiTree;
import nenok.net.Knowledgebase;
import nenok.net.Locator;
import nenok.net.processor.Processor;
import nenok.net.task.TaskFactory;
import nenok.va.Domain;
import nenok.va.Identity;
import nenok.va.Predictability;
import nenok.va.Predictor;
import nenok.va.Scalability;
import nenok.va.VAException;
import nenok.va.Valuation;

/**
 * This is the base class for the various join tree implementations. Each join tree class is equivalent to an architecture 
 * of local computation.
 * 
 * @author Marc Pouly
 * @version 1.1
 */

public abstract class JoinTree implements Iterable<Node> {
    
    // Tree components:
    protected Node root;
    
    // Runtime statistics:
    private double collectTime = -1, distributeTime = -1;
    
    // Predicates:
    private boolean isCollected, isDistributed;
    
    // Knowledgebase:
    private Knowledgebase kb;
    
    // Construction algorithm:
    private Algorithm algorithm;
    
    // Processor allocator:
    protected Allocator allocator;     
    
    // Queries:
    private HashSet<Domain> queries;
   
    // Construction data:
    private Construction cdata;
        
    // Task factory:
    protected TaskFactory factory;
              
    /**
     * Constructor:
     * @param kb The knowledgebase covered by this join tree.
     * @param queries The set of queries that are covered by this join tree.
     * @param algo The construction algorithm.
     * @param proc The processor assigned to query nodes.
     * @throws ConstrException Exceptions thrown during the construction process.
     * @throws RemoteException Exceptions caused by communication problems when accessing the query processor.
     */
    
    protected JoinTree(Knowledgebase kb, Collection<Domain> queries, Algorithm algo, Processor proc) throws ConstrException, RemoteException {
              
    	// Check arguments:
    	if(kb == null || queries == null) {
    		throw new ConstrException("Invalid argument: Either knowledgebase or query argument is null.");
    	}
    	    	
    	if(queries.size() != 0 && proc == null) {
    		throw new ConstrException("Invalid argument: A processor must be specified is the query list is not empty.");    		
    	}
    	    	
        if(!Domain.union(queries).subSetOf(kb.getDomain())) {
            throw new ConstrException("The query variables "+
                    Domain.difference(Domain.union(queries), kb.getDomain())+
                    " are not contained in the knowledgebase.");
        }
        
        /*
         * Test instance acceptability:
         */

        if(!verify(kb.getContentClass())) {
            throw new ConstrException("Cannot use this architecture together with " +
                                      "the given knowledgebase.\nMathematical restrictions are violated.");
        }
        
        this.queries = new HashSet<Domain>(queries);
        
        /*
         * Shrink query list, i.e. eliminate redundant &amp; covered queries:
         */
        
        ArrayList<Domain> minQueries = new ArrayList<Domain>();
        
        // Remove queries that are covered by knowledgebase factors:
        for(Domain query : queries) {
            boolean insert = true;
            for(Locator loc : kb) {
                if(query.subSetOf(loc.getDomain())) {
                    insert = false;
                    break;
                }
            }
            if(insert) {
                minQueries.add(query);
            }
        }
        
        // Remove redundant queries:
        for(int i = 0; i < minQueries.size(); i++) {
            for (int j = i+1; j < minQueries.size(); j++) {
                if(minQueries.get(i).subSetOf(minQueries.get(j))) {
                    minQueries.remove(i--);
                    break;
                }
                if(minQueries.get(j).subSetOf(minQueries.get(i))) { 
                    minQueries.remove(j--);
                }
            }            
        }
                
        /*
         * Knowledgebase:
         */
        
        this.kb = kb;
                
        /*
         * Create task factory:
         */
        
        factory = new TaskFactory();
                
        /*
         * Construction algorithm:
         */
        
        this.algorithm = (algo == null)? new OSLA_SC() : algo;
        
        /*
         * Query processor:
         */
        
        Processor qProc = proc;
        
        if(qProc == null) {
        	for(Locator loc : kb) {
        		qProc = Services.findProcessor(loc);
        		if(qProc != null) {
        			break;
        		}
        	}
        	
        }
        
        this.cdata = new Construction(kb, minQueries, qProc);
        
        /*
         * Launch construction:
         */
        
        double start = System.currentTimeMillis();
        this.root = algorithm.buildTree(cdata);
        this.algorithm.setConstructionTime(System.currentTimeMillis() - start);
    }
    
    /*
     * Abstract methods:
     */
    
    /**
     * Performs the scaling operations for this architecture.
     * This method is executed, if the valuations are type {@link Scalability}.
     * @throws VAException Exception thrown while executing the scaling operation.
     * @throws RemoteException Exceptions caused by communication problems when accessing the query processor.
     */
    
    public abstract void scale() throws VAException, RemoteException;
        
    /**
     * @return The join tree's architecture type.
     */
    
    public abstract String getArchitecture();
        
    /**
     * Verifies that the given knowledgebase can indeed be used with this architecture.
     * @param type The type of the locators that are contained in the knowledgebase.
     * @return <code>true</code>, if the knowledgebase satifies the mathematical restrictions for this architecture.
     */
    
    public abstract boolean verify(Class type);
        
    /*
     * BUILDER design pattern:
     */
        
	/**
	 * @return Returns a new node instance corresponding to the architecture type of this jointree.
	 */
	
	public abstract Node createNode();
	    
	/**
	 * @return The set of queries that are guarantueed to be covered by this jointree.
	 */
	
	public Collection<Domain> getQueries() {
		return queries;
	}
	
    /*
     * Propagation methods:
     */
    
	/**
	 * Executes collect &amp; distribute propagation.
     * @param args <code>args[0] = false</code> averts scaling if the 
     * valuation algebra instance is of type {@link Scalability}.
	 * @throws LCException Exception thrown during join tree propagation.
     * @throws RemoteException Exceptions caused by communication problems when accessing processors.
	 */
	
	public void propagate(boolean... args) throws LCException, RemoteException {
        collect();
		distribute(args);
	}
	
	/**
	 * Executes the collect propagation.
     * @throws LCException Exception thrown during join tree propagation.
     * @throws RemoteException Exceptions caused by communication problems when accessing processors.
	 */
	
	public void collect() throws LCException, RemoteException {
        
        if(isCollected) {
            throw new LCException("Collect has already been executed.");
        }
        
        /*
         * Processor allocation model:
         */
                
        if(allocator == null) {
            if(Utilities.isInterfaceOf(Predictability.class, kb.getContentClass())) {
            	Predictor pred = kb.getPredictor();
            	allocator = new Multiway(getProcessorSet(), pred, root);
            } else {
                allocator = new Default(cdata.proc);
            }
        }
                             
        /*
         * Start collect:
         */
        
        Set<Node> leaves = root.getLeaves();
        double start = System.currentTimeMillis();
		for (Node leaf : leaves) {
            leaf.collect(allocator);
        }
                  
		collectTime = System.currentTimeMillis() - start;
        isCollected = true;
	}
	
	/**
	 * Executes the distribute propagation.
     * @param args <code>args[0] = false</code> averts scaling if the 
     * valuation algebra instance is of type {@link Scalability}.
	 * @throws LCException Exception thrown when trying to execute distribute before collect has been terminated.
     * @throws RemoteException Exceptions caused by communication problems when accessing processors.
	 */
	
	public void distribute(boolean... args) throws LCException, RemoteException {
	    if(!isCollected) {
	        throw new LCException("Collect algorithm must be executed before distribute.");
        }
        
        /*
         * Perform scaling if needed and possible:
         */
   
        if(args.length == 0 || args[0]) {
            Class c = cdata.knowledgebase.getContentClass();
            if(Utilities.isInterfaceOf(Scalability.class, c)) {
                scale();
            }
        }
        
        /*
         * Start distribute:
         */
        
		double start = System.currentTimeMillis();
		root.distribute();
		distributeTime = System.currentTimeMillis() - start;
        isDistributed = true;
	}
	    
	/*
	 * Query answering:
	 */
	
	/**
	 * Method to answer a given query represented by a domain.
	 * @param query The query to answer.
	 * @return The valuation of an (arbitrary) node in the tree with the query's domain.     
     * @throws RemoteException Exceptions caused when trying to access the processor hosting the query answer.
	 */
	
	public Valuation answer(Domain query) throws RemoteException {
        try {
            Node node = root.findCoveringNode(query);
            if(node != null) {
                Valuation val = node.getValuation();
                if(val == null) {
                	return Identity.INSTANCE;
                }
            	Domain dom = Domain.intersection(query, val.label());
            	return val.marginalize(dom);
            }
        } catch (VAException e) {
            e.printStackTrace();
            // cannot happen ...
        }
        return null;
	}
           
    /*
     * Propagation Statistics:
     */
    
    /**
     * @return The time elapsed while executing collect. <code>-1</code>, if collect has not yet been executed.
     */
    
    public double getCollectTime() {
        return collectTime;
    }
    
    /**
     * @return The time elapsed while executing distribute. <code>-1</code>, if distribute has not yet been executed.
     */
    
    public double getDistributeTime() {
        return distributeTime;
    }
    
    /**
     * @return The time elapsed while executing collect and distribute. <code>-1</code>, if the propagation has not yet been executed.
     */
    
    public double getPropagationTime() {
        return collectTime+distributeTime;
    }
    
    /**
     * @return The task factory of this jointree.
     */
    
    public TaskFactory getTaskFactory() {
        return this.factory;
    }
    
    /*
     * Processor allocation:
     */
    
    /**
     * Sets the jointree's processor allocation model.
     * @param allocator The new processor allocation model.
     */
    
    public void setAllocator(Allocator allocator) {
        if(isCollected) {
            throw new UnsupportedOperationException("Cannot change allocation strategy when allocation is already performed.");
        }
        this.allocator = allocator;
    }
    
    /**
     * @return The processor allocation model of this jointree.
     */
    
    public Allocator getAllocator() {
        return allocator;
    }
    
    /**
     * @return Returns the total communication costs caused by local computation messages.
     */
    
    public int getCommunicationCosts() {
        return (allocator == null)? -1 : allocator.totalCosts();
    }
    
    /*
     * Join tree construction:
     */
    
    /**
     * @return The construction algorithm having built this join tree.
     */
    
	public Algorithm getConstructionAlgorithm() {
		return algorithm;
	}
	        
    /*
     * Processor distribution:
     */
    
    /**
     * @return A set of all processors that are assigned to at least one node on this join tree.
     * Note that the null elements will never occur in this set.
     */
    
    public Set<Processor> getProcessorSet() {
        return root.findProcessors();
    }
        
    /*
     * Tree components:
     */
    
    /**
     * @return The root node of this join tree.
     */
    
    public Node getRoot() {
        return this.root;
    }
    
	/**
	 * Jointrees can iterate over their nodes.
	 * @return The jointree's node iterator.
     * @see java.lang.Iterable#iterator()
	 */
    
    public Iterator<Node> iterator() {
    	Set<Node> nodes = root.getNodes(new HashSet<Node>());
    	return nodes.iterator();
    }
    
	/*
	 * Informative methods:
	 */
		
	/**
	 * @return The largest domain in this join tree.
	 */
	
	public Domain getLargestDomain() {
        return root.findLargestDomain();
	}
	    
    /**
     * @return The union of all node domains within the tree.
     */
    
    public Domain getTreeDomain() {
        return root.getTotalDomain();
    }
    
    /**
     * @return Counts the number of nodes within the join tree.
     */
	
    public int countNodes() {
        return root.subTreeSize();
    }
    
    /**
     * @see java.lang.Object#toString()
     */
    
    public String toString() {
    	return getArchitecture();
    }
    
    /**
     * Prints the jointree as ASCII string.
     * @param content <code>true</code> to print the content of each node. 
     * @return The current jointree as ASCII string.
     */
    
    public String toASCII(boolean content) {
    	return AsciiTree.joinTreeToString(root, content);
    }
    
    /*
     * Predicates:
     */
    
    /**
     * @return <code>true</code>, if the collect algorithm has been executed.
     */
    
    public boolean isCollected() {
        return this.isCollected;
    }
    
    /**
     * @return <code>true</code>, if the distribute algorithm has been executed.
     */
    
    public boolean isDistributed() {
        return this.isDistributed;
    }    
    
    /**
     * @return <code>true</code>, if this join tree is binary.
     */
    
    public boolean isBinary() {
        return root.isBinary();
    }
      
    /*
     * Tranformations:
     */
    
    /**
     * Transforms this join tree to a binary join tree.
     */
    
    public void makeBinary() {
        root.makeBinary(cdata);
    }
    
    /*
     * Inner classes:
     */
    
    /**
     * Inner class wrapping all necessary data for the join tree construction process.
     * 
     * @author Marc Pouly
     * @version 1.1
     */
    
    public class Construction {
        
        /**
         * Knowledgebase covered by this join tree.
         */
        
        public Knowledgebase knowledgebase;
        
        /**
         * List of queries to be covered.
         */ 
        
        public List<Domain> queries;
        
        /**
         * Processor assigned to query nodes.
         */
        
        public Processor proc;
        
        /**
         * Constructor:
         * @param kb The knowledgebase covered by this join tree.
         * @param queries List of queries to be covered.
         * @param proc The store of the query processor.
         */
        
        public Construction(Knowledgebase kb, List<Domain> queries, Processor proc) {
            this.knowledgebase = kb;
            this.queries = queries;
            this.proc = proc;
        }
        
        /**
         * @return Returns a {@link Node} instance for the corresponding architecture.
         */
        
        public Node getNodeInstance() {
            return JoinTree.this.createNode();
        }
    }
}