package nenok.lc;

import java.rmi.RemoteException;
import java.util.ArrayList;
import java.util.HashSet;
import java.util.List;
import java.util.Set;

import nenok.Services;
import nenok.lc.alloc.Allocator;
import nenok.net.IdentityLocator;
import nenok.net.Locator;
import nenok.net.processor.Processor;
import nenok.net.task.TaskFactory;
import nenok.va.Domain;
import nenok.va.Valuation;

/**
 * This abstract class represents nodes of a join tree. The tree itself is built from a double linked set of 
 * {@link Node} objects and the {@link JoinTree} object points to the root node of the tree.
 * 
 * @author Marc Pouly
 * @version 1.1
 */

public abstract class Node {
    
    // Unique child node:
    private Node child;
    
    // List of parent nodes:
    private List<Node> parents;
    
    // Node domain:
    private Domain domain;
    
    // Node content locator:
    protected Locator locator;
        
    // Task factory:
    protected TaskFactory factory;
                
    /**
     * Constructor:
     * @param factory The task factory to construct new tasks.
     */
    
    public Node(TaskFactory factory) {
        this.factory = factory;
        parents = new ArrayList<Node>();
    }
           
    /*
     * Propagation methods:
     */
    
    /**
     * Executes the collect algorithm for this node.
     * @param allocator The processor allocation strategy for this node.
     * @throws LCException Exception caused by the collect propagation.
     * @throws RemoteException Exceptions caused by communication problems when accessing processors.
     */
    
    public abstract void collect(Allocator allocator) throws LCException, RemoteException;
    
    /**
     * Executes the distribute algorithm for this node.
     * @throws LCException Exception caused by the distribute propagation.
     * @throws RemoteException Exceptions caused by communication problems when accessing processors.
     */
    
    public abstract void distribute() throws LCException, RemoteException;
    
    /*
     * Tree components:
     */
    
    /**
     * Assigns a unique child element to this node.
     * @param child The node's new child element.
     */
    
    public void setChild(Node child) {
        this.child = child;
    }
    
    /**
     * @return This node's unique child node.
     */
    
    public Node getChild() {
        return child;
    }
    
    /**
     * Adds a new parent node to this node instance.
     * @param parent The new parent node to add.
     */
    
    public void addParent(Node parent) {
        if(!parents.contains(parent)) {
            parents.add(parent);
        }
        parent.child = this;
    }
    
    /**
     * Adds a new parent node to this node instance at the given position.
     * @param parent The new parent node to add.
     * @param index The position where the new parent is added.
     */
    
    private void addParent(Node parent, int index) {
        if(!parents.contains(parent)) {
            parents.add(index, parent);
        }
        parent.child = this;        
    }
    
    /**
     * Adds new new parent nodes to this node instance.
     * @param parents A list of new parent nodes to add.
     */
    
    public void addParents(List<Node> parents) {
        for(Node parent : parents) {
            this.addParent(parent);
        }
    }
        
    /**
     * @return The parent nodes of this node instance.
     */
    
    public List<Node> getParents() {
        return parents;
    }
    
    /**
     * Removes a parent node from this node instance.
     * @param parent The parent to remove.
     */
    
    protected void removeParent(Node parent) {
        parents.remove(parent);  
    }
        
    /**
     * @return The leaf nodes in the subtree of this node.
     */
    
    public Set<Node> getLeaves() {
        return findLeaves(new HashSet<Node>());
    }
    
    /*
     * Node content:
     */
    
    /**
     * @return Returns the domain of this node.
     */
    
    public Domain getDomain() {
        return domain;
    }
    
    /**
     * Sets the domain of this node.
     * @param domain The domain to set.
     */
    
    public void setDomain(Domain domain) {
        this.domain = domain;
    }
    
    /**
     * @return Returns the node content of this node.
     * @throws RemoteException Exception thrown while reading the node content from the remote processor. 
     */
    
    public Valuation getValuation() throws RemoteException {
        return (locator == null) ? null : locator.retrieve();
    }
    
    /**
     * Sets the content locator of this node instance.
     * @param loc The locator pointing to the content of this node.
     */
    
    public void setLocator(Locator loc) {
        this.locator = loc;
    }
    
    /**
     * @return The locator that points to this node's valuation.
     */
    
    public Locator getLocator() {
        return this.locator;
    }
    
    /**
     * @return The processor assigned to this node.
     */
    
    public Processor getProcessor() {
        return (locator == null)? null : Services.findProcessor(locator);
    }
    
    /**
     * Assigns a processor to this node. The method's behaviour is:
     * <ul>
     * <li>If no locator is assigned to this node, a new identity element is created on the given processor 
     * and the locator for it is assigned to this node. The return value is <code>true</code>.</li>
     * <li>If this node already possesses a locator, nothing happens and <code>false</code> is returned.</li>
     * </ul>
     * @param proc The processor to assign.
     * @return <code>true</code>, if the processor assigned to this nodes changes.
     * @throws RemoteException Exception caused by communication proclems when the new processor is accessed.
     */
    
    public boolean setProcessor(Processor proc) throws RemoteException {
        if(locator == null) {
        	setLocator(new IdentityLocator(proc.getPID()));
            return true;
        }
        return false;
    }
    
    /*
     * Informative methods:
     */
	
    /**
     * @return The number of nodes in the subtree of this node.
     */
	
    public int subTreeSize() {
        int result = 1;
        for(Node node : parents) {
            result += node.subTreeSize();
        }
        return result;
    }
    
    /**
     * @return <code>true</code>, if this node's subtree is binary.
     */
    
    protected boolean isBinary() {
        boolean result = parents.size() <= 2;
        for (Node parent : parents) {
            result &= parent.isBinary();
        }
        return result;
    }
    
    /**
     * @return The largest node domain within the subtree of this node.
     */
    
    protected Domain findLargestDomain() {
        Domain result = domain;
        for (Node parent : parents) {
            Domain pdom = parent.findLargestDomain();
            if(pdom.size() > result.size()) {
                result = pdom;
            }            
        }
        return result;
    }
    
    /**
     * @return The union of all node domains within the subtree of this node.
     */
    
    protected Domain getTotalDomain() {
        Domain result = domain;
        for(Node node : parents) {
            result = Domain.union(result, node.getTotalDomain());
        }
        return result;
    }
      
    /**
     * @return The set of all processors assigned in the subtree of this node.
     */
    
    public Set<Processor> findProcessors() {
        return findProcessors(new HashSet<Processor>());
    }
    
    /*
     * Query answering:
     */
    
    /**
     * @param query The query domain to find.
     * @return A node whose domain equals the query or <code>null</code>.
     */
    
    protected Node findCoveringNode(Domain query) {
      if(query.subSetOf(domain)) {
          return this;
      }
      for (Node parent : parents) {
          Node result = parent.findCoveringNode(query);
          if(result != null) {
              return result;
          }
      }
      return null;
    }

    /*
     * Tree transformations:
     */
    
    /**
     * Transforms the subtree of this to a binary tree.
     * @param data The information for the join tree construction process.
     */
    
    protected void makeBinary(JoinTree.Construction data) {
        
        /*
         * Iterate as long as the current node is not binary.
         */
        
        while(getParents().size() > 2) {
            Node first = parents.remove(0);
            Node second = parents.remove(1);
            Node union = data.getNodeInstance();
            union.setDomain(Domain.intersection(Domain.union(first.domain, second.domain), getDomain()));
            union.addParent(first);
            union.addParent(second);
            addParent(union, parents.size() - 1);
        }
        
        /*
         * Apply this procedure to all parent nodes.
         */
        
        for(Node parent : parents) {
            parent.makeBinary(data);
        }
    }

    /*
     * Helper mthods:
     */
    
    protected Set<Node> getNodes(Set<Node> nodes) {
    	nodes.add(this);
    	for(Node parent : parents) {
    		parent.getNodes(nodes);
    	}
    	return nodes;
    }
        
    /**
     * Helper procedure to find all leaves of this node's subtree.
     * @param cache The cache to store leaves.
     * @return The filled cache.
     */
    
    private Set<Node> findLeaves(Set<Node> cache) {
        if(parents.size() == 0) {
            cache.add(this);
            return cache;
        }
        for (Node parent : parents) {
            parent.findLeaves(cache);
        }
        return cache;
    }
    
    /**
     * Helper procedure to find all processors in this node's subtree.
     * @param processors The cache to store processors.
     * @return The filled cache.
     */
    
    private Set<Processor> findProcessors(Set<Processor> processors) {
    	Processor p = this.getProcessor();
    	if(p != null) {
    		processors.add(p); 
    	}           
        for(Node parent : parents) {
            parent.findProcessors(processors);
        }
        return processors;
    }
}