package nenok.lc.alloc;

import java.rmi.RemoteException;
import java.util.Hashtable;

import nenok.lc.Node;
import nenok.net.processor.Processor;

/**
 * This is a default implementation of a dynamic processor allocation model.
 * To determine the processor to allocation to a given node, this model only
 * considers the allocation of the node's parents.
 * 
 * @author Marc Pouly
 * @version 1.1
 */

public class Default implements Allocator {
    
    private int costs;
    private Processor proc;
    
    /**
     * Constructor:
     * @param proc The processor that is allocated to leaf nodes.
     */
    
    public Default(Processor proc) {
        this.proc = proc;
    }

    /**
     * @see nenok.lc.alloc.Allocator#allocate(nenok.lc.Node)
     */
    
    public void allocate(Node node) throws RemoteException {
               
        if(node.getProcessor() != null) {
            return;
        }
        
        if(node.getParents().size() == 0) {
            node.setProcessor(proc);
            return;
        }
        
        /*
         * Compute economization:
         */

        Hashtable<Processor, Double> table = new Hashtable<Processor, Double>();
        
        for(Node parent : node.getParents()) {
            if(!table.containsKey(parent.getProcessor())) {
                table.put(parent.getProcessor(), 0.0);
            }
            double result = table.get(parent.getProcessor());
            result += parent.getLocator().getWeight();
            table.put(parent.getProcessor(), result);
        }
        
        /*
         * Search maximum value:
         */
        
        double max = 0;
        for(Double d : table.values()) {
            if(d >= max) {
                max = d;
            }
        }
        
        for(Processor p : table.keySet()) {
            if(table.get(p) == max) {
                node.setProcessor(p);
            }
        }
        
        /*
         * Communication costs:
         */
        
        for(Double value : table.values()) {
            costs += value;
        }
        costs -= max;
    }
    
    /**
     * @see nenok.lc.alloc.Allocator#getName()
     */
    
    public String getName() {
        return "Dynamic Allocation";
    }
    
    /**
     * @see nenok.lc.alloc.Allocator#totalCosts()
     */
    
    public int totalCosts() {
        return costs;
    }
}
