package nenok.lc.alloc;

import java.rmi.RemoteException;
import java.util.ArrayList;
import java.util.Hashtable;
import java.util.List;
import java.util.Set;

import nenok.lc.Node;
import nenok.net.processor.Processor;
import nenok.va.Domain;
import nenok.va.Predictor;

/**
 * This class performs processor allocation to join tree node by 
 * use of the minimization algorithm for multiway cuts on trees.
 * Basically, it is a static allocation model which means that 
 * processor allocation is fully performed before collect is started.
 * 
 * @author Marc Pouly
 * @version 1.1
 */

public class Multiway implements Allocator {
    
    private int costs;
    private Predictor pred;
    private List<Processor> procs;
    private Hashtable<Node, Penalty> ndata;
    
    /**
     * Constructor:
     * @param procs The allocatable processors.
     * @param pred The weight predictor for the communication costs.
     * @param root The root node of the jointree.
     * @throws RemoteException Exception caused by accessing the processors that are assigned.
     */
        
    public Multiway(Set<Processor> procs, Predictor pred, Node root) throws RemoteException {
        procs.remove(null);
        
        this.pred = pred;
        this.procs = new ArrayList<Processor>(procs);
        this.ndata = new Hashtable<Node, Penalty>(); 
        
        /*
         * Start inward phase:
         */
        
        Set<Node> leaves = root.getLeaves();
        for(Node leaf : leaves) {
            inward(leaf);
        }
        
        /*
         * Allocator root processor:
         */
        
        Penalty penalty = ndata.get(root);
        root.setProcessor(penalty.getProcessorWithMinPenalty());
        
        /*
         * Start outward phase:
         */
        
        for(Node parent : root.getParents()) {
            outward(parent);
        }
    }
    
    /**
     * @see nenok.lc.alloc.Allocator#allocate(nenok.lc.Node)
     */
    
    public void allocate(Node node) throws RemoteException {
        
        /*
         * Processor is already assigned - only real costs are computed.
         */
        
        if(node.getChild() != null) {
            
            if(node.getChild().getProcessor() == null) {
                System.out.println("Child proc is null !!!");
                return;
            }
            
            if(!node.getChild().getProcessor().getPID().equals(node.getProcessor().getPID())) {
                costs += node.getChild().getLocator().getWeight();
                costs += pred.predict(Domain.intersection(node.getChild().getDomain(), node.getDomain()));
            }
        }
    }
        
    /**
     * Inward phase: Computes the penalty values for each node.
     * @param node The current node in the recursive run of the inward phase.
     */
    
    private void inward(Node node) {
        for(Node parent : node.getParents()) {
            if(!ndata.containsKey(parent)) {
                return;
            }
        }
        ndata.put(node, new Penalty(node));
        if(node.getChild() != null) {
            inward(node.getChild());
        }
    }
    
    /**
     * Outward phase: Assigns a processor to each node.
     * @param node The current node in the recursive run of the outward phase.
     * @throws RemoteException Exception caused by accessing the processors that are assigned.
     */
    
    private void outward(Node node) throws RemoteException {
        
        if(node.getProcessor() == null) {
            double min = Double.MAX_VALUE;
            int index = -1;
            for (int i = 0; i < procs.size(); i++) {
                double weight = 0;
                if(!node.getChild().getProcessor().equals(procs.get(i))) {
                    weight += costs;
                }
                weight += ndata.get(node).valueOf(procs.get(i));
                if(weight < min) {
                    min = weight;
                    index = i;
                }
            }
            node.setProcessor(procs.get(index));
        } 
        
        for(Node parent : node.getParents()) {
            outward(parent);
        }
    }
    
    /**
     * @see nenok.lc.alloc.Allocator#getName()
     */
    
    public String getName() {
        return "Multiway";
    }
    
    /**
     * @see nenok.lc.alloc.Allocator#totalCosts()
     */
    
    public int totalCosts() {
        return costs;
    }
    
    /**
     * Implementation of the penalty function.
     * 
     * @author Marc Pouly
     * @version 1.1
     */
    
    private class Penalty {
                
        /*
         * Penalty function: Processor -> double
         */
        
        private double[] penalties;
        
        /**
         * Constructor:
         * @param node The node whose penalty values are computed.
         */
        
        private Penalty(Node node) {
            
            penalties = new double[procs.size()];
            
            /*
             * Leaves without assigned processor:
             */
            
            if(node.getParents().size() == 0 && node.getProcessor() == null) {
                for (int i = 0; i < penalties.length; i++) {
                    penalties[i] = 0;
                }
            }
            
            /*
             * Nodes with assigned processor:
             */
            
            else if(node.getProcessor() != null) {
                for(int i = 0; i < penalties.length; i++) {
                    if(procs.indexOf(node.getProcessor()) == i) {
                        penalties[i] = 0;
                    } else {
                        penalties[i] = Double.MAX_VALUE;
                    }
                }
                return;
            }
            
            /*
             * Intermediate nodes without assigned processor:
             */
            
            else {
                for (int i = 0; i < penalties.length; i++) {
                    double result = 0;
                    for(Node parent : node.getParents()) {
                        Penalty pData = ndata.get(parent);
                        double min = Double.MAX_VALUE;
                        for(int j = 0; j < procs.size(); j++) {
                            double penalty = pData.valueOf(procs.get(j));
                            if(i != j) {
                                penalty += pred.predict(Domain.intersection(parent.getDomain(), node.getDomain()));
                            }
                            if(penalty < min) {
                                min = penalty;
                            }
                        }
                        result += min;
                    }
                    penalties[i] = result;
                }
                return;
            }
        }
        
        /**
         * @param proc A processor
         * @return The penalty if processor <code>proc</code> is assigned to the current node.
         */
        
        private double valueOf(Processor proc) {
            return penalties[procs.indexOf(proc)];
        }
    
        /**
         * @return The minimum penalty value.
         */
        
        private Processor getProcessorWithMinPenalty() {
            int index = -1;
            double result = Double.MAX_VALUE;
            for (int i = 0; i < penalties.length; i++) {
                if(penalties[i] < result) {
                    result = penalties[i];
                    index = i;
                }
            }
            return procs.get(index);
        }
    }
}
