package nenok.lc.hugin;

import java.rmi.RemoteException;

import nenok.lc.LCException;
import nenok.lc.Node;
import nenok.lc.alloc.Allocator;
import nenok.net.Locator;
import nenok.net.task.Task;
import nenok.net.task.TaskFactory;
import nenok.va.Domain;
import nenok.va.VAException;

/**
 * This is the node class implementation that belongs to the HUGIN architecture.
 * 
 * @author Marc Pouly
 * @version 1.2
 */

public class HNode extends Node {

    // Internal message cache:
	private int counter;
    private Locator[] messages;
    
    // Separator:
    private Locator separator;
    
    /**
     * Constructor:
     * @param factory The task factory for this node.
     */
    
    public HNode(TaskFactory factory) {
        super(factory);
    }
                
    /**
     * @see nenok.lc.Node#collect(nenok.lc.alloc.Allocator)
     */
    
    public void collect(Allocator allocator) throws LCException, RemoteException { 
        
        // Test, if all parents have sent their message:
        if(counter < this.getParents().size()) {
            return;
        }
        
        // Allocate processor -> locator != null after this call.
        allocator.allocate(this);
        
        // Compute new node content -> messages haven't been initialized for leaves.  
        if(getParents().size() > 0) {
            messages[counter] = locator;
            Task task = factory.createCombination(messages);
            locator = this.getProcessor().execute(task); 
                    
            // Throw messages away:
            messages = null;
        }
                                   
        // Stop when reaching the root node:
        if(getChild() == null) {
            return;
        }
                      
        // Compute message for the child:
        Domain inter = Domain.intersection(locator.getDomain(), getChild().getDomain());  
        Task task = factory.createMarginalization(locator, inter);
        separator = this.getProcessor().execute(task);
                        
        // Send message to child:
        HNode child = (HNode)this.getChild();
        child.receiveParentMessage(separator);
        child.collect(allocator);
    }

    /**
     * @see nenok.lc.Node#distribute()
     */
    
    public void distribute() throws LCException, RemoteException {
        for(Node parent : this.getParents()) {
            Task task = factory.createMarginalization(locator, Domain.intersection(this.getDomain(), parent.getDomain()));
            Locator message = this.getProcessor().execute(task);
        	((HNode)parent).receiveChildMessage(message);
            parent.distribute();
        }
    }
   
    /*
     * Helper methods:
     */
    
    /**
     * Method called by parent node to send message to this node.
     * @param message The message to send.
     */
    
    private void receiveParentMessage(Locator message) {

    	// Initialization of message cache.
        if(messages == null) {
            messages = new Locator[getParents().size()+1];
        }
        
    	messages[counter++] = message;
    }
    
    /**
     * Method called by child node to send message to this node.
     * @param message The message to send.
     * @throws RemoteException Exceptions caused by communication problems when accessing processors.
     */
    
    private void receiveChildMessage(Locator message) throws RemoteException {
        try {
            
            // Inverse of separator:
            Task inv = factory.createInverse(separator);
            Locator inverse = this.getProcessor().execute(inv);
                        
            // Combine message with inverse of separator:
            Task comb = factory.createCombination(inverse, message);
            Locator idempotent = this.getProcessor().execute(comb);
            
            // Combine idempotent to the node content:
            comb = factory.createCombination(idempotent, locator);
            locator = this.getProcessor().execute(comb);
                        
        } catch (VAException e) {
            e.printStackTrace();
            // cannot happen ...
        }
     }
}
