package nenok.lc.ss;

import java.rmi.RemoteException;
import java.util.LinkedHashMap;
import java.util.Map;

import nenok.lc.LCException;
import nenok.lc.Node;
import nenok.lc.alloc.Allocator;
import nenok.net.Locator;
import nenok.net.task.Task;
import nenok.net.task.TaskFactory;
import nenok.va.Domain;
import nenok.va.VAException;

/**
 * This is the node class implementation that belongs to the Shenoy-Shafer architecture.
 * Internally, this class stores messages within a {@link LinkedHashMap} in order to ensure
 * that the combination order equals the order messages arrived.
 * 
 * @author Marc Pouly
 * @version 1.2
 */

public class SSNode extends Node {

    // Internal message cache:
    private LinkedHashMap<Node, Locator> messages;
        
    /**
     * Constructor:
     * @param factory The task factory for this node.
     */
    
    public SSNode(TaskFactory factory) {
        super(factory);
        this.messages = new LinkedHashMap<Node, Locator>(this.getParents().size());
    }
    
    /**
     * @see nenok.lc.Node#collect(nenok.lc.alloc.Allocator)
     */
    
    public void collect(Allocator allocator) throws LCException, RemoteException { 
        
        // Test, if all parents have sent their message:
        if(messages.keySet().size() < this.getParents().size()) {
            return;
        }
        
        // Allocate processor:
        allocator.allocate(this);
        
        // Store initial node content:
        messages.put(this, locator);
                
        // Combine messages from parents:
        Locator result = locator;
        for(Map.Entry<Node, Locator> entry : messages.entrySet()) {
            if(!entry.getKey().equals(this)) {
                Task task = factory.createCombination(result, entry.getValue());
                result = this.getProcessor().execute(task);   
            }
        }
        locator = result;
                          
        // Stop when reaching the root node:
        if(getChild() == null) {
            return;
        }
               
        // Compute message for the child:
        Domain dom = Domain.intersection(locator.getDomain(), getChild().getDomain());        
        Task task = factory.createMarginalization(locator, dom);
        Locator message = this.getProcessor().execute(task);
                        
        // Send message to child:
        SSNode child = (SSNode)this.getChild();
        child.receiveParentMessage(message, this);
        child.collect(allocator);
    }

    /**
     * @see nenok.lc.Node#distribute()
     */
    
    public void distribute() throws LCException, RemoteException {
        for(Node parent : this.getParents()) {
            Locator message = computeParentMessage(parent);
            ((SSNode)parent).receiveChildMessage(message);
            parent.distribute();
        }
        
        // Throw messages away:
        messages = null;
    }
    
    /**
     * Helper method to set the scaling factor to the root node.
     * @param factor The scaling factor.
     * @throws RemoteException Exceptions caused by communication problems when accessing processors.
     */
    
    protected void setScalingFactor(Locator factor) throws RemoteException {
        try {
            Locator loc = messages.remove(this);
            Task task = factory.createCombination(factor, loc);
            messages.put(this, this.getProcessor().execute(task));
        } catch (VAException e) {
            e.printStackTrace();
            // cannot happen ...
        }
    }
    
    /*
     * Helper methods:
     */
    
    /**
     * Method called by parent node to send message to this node.
     * @param message The message to send.
     * @param parent The addresser node.
     */
    
    private void receiveParentMessage(Locator message, Node parent){
        messages.put(parent, message);
     }

    /**
     * Method called by child node to send message to this node.
     * @param message The message to send.
     * @throws RemoteException Exceptions caused by communication problems when accessing processors.
     */
    
    private void receiveChildMessage(Locator message) throws RemoteException {
        try {
            messages.put(this.getChild(), message);
            Task task = factory.createCombination(locator, message);
            locator = this.getProcessor().execute(task);
        } catch (VAException e) {
            e.printStackTrace();
            // cannot happen ...
        }
    }
    
    /**
     * Computes the message sent to the given parent.
     * @param parent The parent that will receive the computed message.
     * @return The message for the given parent node.
     * @throws RemoteException Exceptions caused by communication problems when accessing processors.
     */
    
    private Locator computeParentMessage(Node parent) throws RemoteException {
        try {
            Locator result = messages.get(this);
            for(Map.Entry<Node, Locator> entry : messages.entrySet()) {
                if(!entry.getKey().equals(this) && !entry.getKey().equals(parent)) {
                    Task task = factory.createCombination(result, entry.getValue());
                    result = this.getProcessor().execute(task);                
                }
            }
            Task task = factory.createMarginalization(result, Domain.intersection(result.getDomain(), parent.getDomain()));
            return this.getProcessor().execute(task);
        } catch (VAException e) {
            e.printStackTrace();
            // cannot happen ...
        }
        return null;
    }
}
