package nenok.net;

import java.io.Serializable;
import java.net.MalformedURLException;
import java.net.URL;
import java.net.URLClassLoader;
import java.rmi.RemoteException;
import java.util.Arrays;
import java.util.Collection;
import java.util.HashSet;
import java.util.Iterator;

import nenok.Services;
import nenok.Utilities;
import nenok.net.processor.Processor;
import nenok.va.Domain;
import nenok.va.Identity;
import nenok.va.Predictability;
import nenok.va.Predictor;

/**
 * A knowledgebase is basically a set of {@link Locator} objects. Because each locator identifies unambigously a valuation 
 * object stored on a Nenok processor, this represents indeed the mathematical idea of a distributed knowledgebase.
 * <br/><br/>
 * Knowledgebases contain only elements of the same type. However, because {@link Locator} objects are only pointer on the 
 * storage place (processor) of a valuation, this type checking cannot be done at compile-time. So, the first added element 
 * determines which type this knowledgebase accepts. If this constraint is not satisfied, an {@link IllegalArgumentException} is thrown.
 * 
 * @author Marc Pouly
 * @version 1.2
 */

public class Knowledgebase implements Serializable, Iterable<Locator> {
    
    private String type, name;
    private HashSet<Locator> locators;
    
	private volatile Class cache;
    
    /**
     * Constructor:
     */
    
    public Knowledgebase() {
    	name = "Unnamed Knowledgebase";
        locators = new HashSet<Locator>();
    }
        
    /**
     * Constructor:
     * @param locators The locators to add to the new knowledgebase.
     */
        
    public Knowledgebase(Locator[] locators) {
        this();
        addAll(Arrays.asList(locators));
    }
            
    /**
     * Adds a new locator to this knowledgebase.
     * @param loc The locator to add.
     * @return <code>true</code>, it the locator has been added successfully.
     */
    
    public boolean add(Locator loc) {
        if(loc.getType().equals(Identity.class.getName())) {
            return locators.add(loc);
        }
        if(type == null || type == "") {
            type = loc.getType();
            return locators.add(loc);
        }
        if(this.type.equals(loc.getType())) {
            return locators.add(loc);
        }
        return false;
    }
    
    /**
     * Adds a new collection of locators to this knowledgebase.
     * @param locators The locators to add.
     * @return <code>true</code>, it all locators have been added successfully.
     */
    
    public boolean addAll(Collection<Locator> locators) {
        boolean result = true;
        for(Locator l : locators) {
            result &= this.add(l);
        }
        return result;
    }
    
    /**
     * Adds a new array of locators to this knowledgebase.
     * @param locators The locators to add.
     * @return <code>true</code>, it all locators have been added successfully.
     */
    
    public boolean addAll(Locator[] locators) {
        boolean result = true;
        for(Locator l : locators) {
            result &= this.add(l);
        }
        return result;
    }    
    
    /**
     * Removes a given locator from this knowledgebase.
     * @param loc The locator to remove.
     * @return <code>true</code>, if the locator has been removed successfully.
     */
    
    public boolean remove(Locator loc) {
        return locators.remove(loc);
    }
    
    /**
     * Looks for a given locator within this knowledgebase.
     * @param loc The locator to search.
     * @return <code>true</code>, if the locator has been found.
     */
    
    public boolean contains(Locator loc) {
        return locators.contains(loc);
    }
        
    /**
     * @return The size of this knowledgebase.
     */
    
    public int size() {
        return locators.size();
    }
    
    /**
     * @return The knowledgebase transformed into an array of locators.
     */
    
    public Locator[] toArray() {
    	return locators.toArray(new Locator[locators.size()]);
    }
    
    /**
     * Sets the name of this knowledgebase.
     * @param name The name of the knoelwedgebase.
     */
    
    public void setName(String name) {
    	this.name = name;
    }
    
    /**
     * @return The name of this knowledgebase.
     */
    
    public String getName() {
    	return name;
    }
    
    /*
     * Iterable interface:
     */
    
    /**
     * @return The iterator for this knowledgebase.
     * @see java.lang.Iterable#iterator()
     */
    
    public Iterator<Locator> iterator() {
        return locators.iterator();
    }
    
    /*
     * Misc:
     */
    
    /**
     * @see java.lang.Object#toString()
     */
    
    public String toString() {
    	return name;
    }
    
    /**
     * @return The union domain of all locators' domains within this knowledgebase.
     */
    
    public Domain getDomain() {
    	return Domain.totalDomain(locators);
    }
        
    /**
     * @return The predictor of this knowledgebase of null, if it doesn't exist.
     * @throws RemoteException This operation performs remote calls and therefore forewards RemoteExceptions.
     */
    
    public Predictor getPredictor() throws RemoteException {
    	if(Utilities.isInterfaceOf(Predictability.class, this.getContentClass())) {
        	for(Locator loc : locators) {
        		if(loc.getType().equals(type)) {
        			return ((Predictability)loc.retrieve()).predictor();
        		}
        	}
    	}
    	return null;
    }
    
    /**
     * Loads dynamically the class type of the valuations referenced by this knowledgebase.
     * @return The class of the valuations referenced by the locators within this knowledgebase. 
     * Returns <code>null</code>, if this knowledgebase is empty.
     * @throws RemoteException This operation performs remote calls and therefore forewards RemoteExceptions.
     */
    
    public Class getContentClass() throws RemoteException {
    	if(cache != null) {
    		return cache;
    	}
    	reload();
    	return cache;
    }
    
    /**
     * Loads dynamically the class type of the valuations referenced by this knowledgebase.
     * @throws RemoteException This operation performs remote calls and therefore forewards RemoteExceptions.
     */
    
    private void reload() throws RemoteException {
        if(type != null && type != "") {
            for(Locator loc : locators) {
                if(this.type.equals(loc.getType())) {
                    try {
                    	
                        Processor p = Services.findProcessor(loc);
                        
                        /*
                         * For local processors:
                         */
                                               
                        if(p.getCodebase() == null) {
                        	
                        	// With system class loader:
                        	ClassLoader loader = ClassLoader.getSystemClassLoader();
                        	cache = loader.loadClass(type);
                        	return;
                        	
                        	/*if(cache != null) {
                        		return;
                        	}
                        	
                        	// With JAR loader:
                        	cache = Utilities.loadInstance(type);
                        	return;*/
                        
                        }
                                                
                        /*
                         * For remote processors:
                         */

                        String[] str = p.getCodebase().split(" ");                        
                        URL[] codebase = new URL[str.length];
                        for (int i = 0; i < codebase.length; i++) {
                            codebase[i] = new URL(str[i]);
                        }    
                        URLClassLoader loader = new URLClassLoader(codebase);
                        cache =  loader.loadClass(type);
                        return;
                    } catch (ClassNotFoundException e) {
                        throw new RemoteException("Unable to locate valuation's class from processor codebase.", e);
                    } catch (MalformedURLException e) {
                        throw new RemoteException("Malformed URL received from processor.", e);
                    }
                }
            }            
        }
        cache = Identity.class;
    }
}
