package nenok.net;

import java.io.Serializable;
import java.rmi.RemoteException;

import nenok.va.Domain;
import nenok.va.Valuation;
import net.jini.id.Uuid;

/**
 * The location of a (remote) {@link Valuation} object can unambiguously be determined by 
 * the identifier of the processor hosting the valuation and the identifier of the valuation's 
 * envelope. Locators are objects that carries these informations in order to localize valuation 
 * objects at a later time. Furthermore, they manage the following additional informations:
 * <br/><br/>
 * <ul>
 * 		<li>The domain of the referenced valuation. Jointree construction algorithms need to know 
 * 			the domains of the valuations to build jointrees without transmitting valuation objects 
 * 			over the network.
 * 		</li>
 * 		<li>The weight of the referenced valuation. It is used as the principal measure for the processor 
 * 			allocation process and needs therefore to be avaiable at this time. 
 * 		</li>  
 * 		<li>Locator objects are collected within knowledgebases. Such knowledgebases should only contain 
 * 			valuations of the same type. Because only references to valuations are stored in knowledgebases, 
 * 			this cannot be garantueed at compile-time. Each locator contains therefore the package and classname 
 * 			(i.e. <code>v.getClass().getName()</code>) for the valuation object it references. Using this information, 
 * 			knowledgebases can identify the valuation types.
 * 		</li>
 * </ul>
 * 
 * @author Marc Pouly
 * @version 1.3
 */

public abstract class Locator implements Serializable {
	
	private Uuid pid;
	
	/**
	 * Constructor:
     * @param pid The identifier of the processor hosting the valuation.
	 */
	
	protected Locator(Uuid pid) {
		if(pid == null) {
			throw new IllegalArgumentException("Null is not a valid processor identifier.");
		}
		this.pid = pid;
	}
	
    /**
     * @return Retrieves the referenced valuation from the processor.
	 * @throws RemoteException Exception caused when trying to access processors.
     */
	
	public abstract Valuation retrieve() throws RemoteException;
	
	/**
	 * Deletes the valuation that is referenced by this locator.
	 * @throws RemoteException Exception caused when trying to access processors.
	 */
	
	public abstract void delete() throws RemoteException;
	
    /**
     * @return The domain of the referenced valuation.
     */
	
    public abstract Domain getDomain();
	
    /**
     * @return The weight of the referenced valuation.
     */

    public abstract int getWeight();
	
    /**
     * @return The type (class name) of the referenced valuation.
     */
    
    public abstract String getType();
    
    /**
     * @return The indentifier of the processor that hosts the valuation.
     */
    
    public Uuid getPID() {
    	return this.pid;
    }
}
