package nenok.net.processor;

import java.rmi.Remote;
import java.rmi.RemoteException;

import nenok.net.DefaultLocator;
import nenok.net.Envelope;
import nenok.net.Locator;
import nenok.net.task.Task;
import nenok.va.VAException;
import nenok.va.Valuation;
import net.jini.id.Uuid;

/**
 * The interface of a processor in a Nenok federation.
 * 
 * @author Marc Pouly
 * @version 1.2
 */

public interface Processor extends Remote {

    /**
     * Stores a valuation wrapped in an envelope on this processor.
     * @param env The envelope with the valuation to store.
     * @return The locator of the stored valuation (envelope).
     * @throws RemoteException Locator methods all throw remote exceptions.
     */
    
    public Locator store(Envelope env) throws RemoteException;
    
    /**
     * Stores a valuation on this processor.
     * @param val The valuation to store.
     * @return The locator of the stored valuation.
     * @throws RemoteException Locator methods all throw remote exceptions.
     */
    
    public Locator store(Valuation val) throws RemoteException;
    
    /**
     * Stores a set of valuations on this processor.
     * @param vals The valuations to store.
     * @return The locators of the stored valuations. Locator i references valuation i.
     * @throws RemoteException Locator methods all throw remote exceptions.
     */
    
    public Locator[] store(Valuation[] vals) throws RemoteException;
        
    /**
     * Retrieves a stored valuation from this processor.
     * This method call does not remove the corresponding locator from the internal store.
     * @param loc The locator identifying the valuation.
     * @return The valuation referenced by the locator or null.
     * @throws RemoteException Locator methods all throw remote exceptions.
     */
    
    public Valuation retrieve(DefaultLocator loc) throws RemoteException; 
    
    /**
     * Is the valuation with the given envelope ID stored on this processor.
     * @param envID The valuation's envelope ID.
     * @return <code>true</code>, if the valuation exists.
     * @throws RemoteException Locator methods all throw remote exceptions.
     */
    
    public boolean exists(Uuid envID) throws RemoteException;
    
    /**
     * Deletes a stored valuation from this processor.
     * @param loc The locator referencing the valuation to delete.
     * @throws RemoteException Locator methods all throw remote exceptions.
     */
    
    public void delete(DefaultLocator loc) throws RemoteException;
    
    /**
     * Deletes all stored valuations.
     * @throws RemoteException Locator methods all throw remote exceptions.
     */
    
    public void reset() throws RemoteException; 
    
    /*
     * Task processing:
     */
    
    /**
     * Executes a given task on this processor. Tasks are encapsulated valuation algebra operations (COMMAND DESIGN PATTERN).
     * @param task The task to execute.
     * @return The locator of the task's result.
     * @throws RemoteException Locator methods all throw remote exceptions.
     * @throws VAException Exception caused by undefined valuation algera operations.
     */
    
    public Locator execute(Task task) throws RemoteException, VAException;
    
    /*
     * Informational methods:
     */
        
    /**
     * @return The identifier of this processor.
     * @throws RemoteException Locator methods all throw remote exceptions.
     */
    
    public Uuid getPID() throws RemoteException;

    /**
     * @return The codebase of this processor at start time.
     * @throws RemoteException Locator methods all throw remote exceptions.
     */
    
    public String getCodebase() throws RemoteException;
}