package nenok.net.processor;

import java.rmi.RemoteException;
import java.util.concurrent.ConcurrentHashMap;
import java.util.logging.Level;
import java.util.logging.Logger;

import nenok.net.DefaultLocator;
import nenok.net.Envelope;
import nenok.net.Locator;
import nenok.net.task.Task;
import nenok.va.VAException;
import nenok.va.Valuation;
import net.jini.id.Uuid;

/**
 * An implementation of the {@link Processor} interface.
 * 
 * @author Marc Pouly
 * @version 1.2
 */

public class ProcessorImpl implements Processor {
    
    private final Uuid pid;
    private String codebase;
    private final ConcurrentHashMap<Uuid, Envelope> table;
    
    /*
     * Ensure that a SecurityManager is running.
     */
    
    private static Logger logger;
    
    static {
        logger = Logger.getLogger(Processor.class.getPackage().toString());
        logger.setLevel(Level.WARNING);
    }
                
    /**
     * Constructor:
     * @param pid The identifier of the processor.
     */
    
    public ProcessorImpl(Uuid pid) {
        this.pid = pid;
        this.table = new ConcurrentHashMap<Uuid, Envelope>(); 
    	try {
    		this.codebase = System.getProperty("java.rmi.server.codebase");
    	} catch(Exception e) {
    		System.err.println("Exception from Security-Manager: Probably no policy file defined.");
    		e.printStackTrace();
    		System.exit(1);
    	}
    }
    
    /**
     * @see nenok.net.processor.Processor#store(nenok.va.Valuation)
     */
    
    public Locator store(Valuation val) throws RemoteException {
        Envelope env = new Envelope(val);
        table.put(env.getID(), env);
        return new DefaultLocator(pid, env);
    }
    
    /**
     * @see nenok.net.processor.Processor#store(nenok.net.Envelope)
     */
    
    public Locator store(Envelope env) throws RemoteException {
        table.put(env.getID(), env);
        return new DefaultLocator(pid, env);
    }

    /**
     * @see nenok.net.processor.Processor#store(nenok.va.Valuation[])
     */
    
    public Locator[] store(Valuation[] vals) throws RemoteException {
        Locator[] result = new Locator[vals.length];
        for (int i = 0; i < vals.length; i++) {
            result[i] = this.store(vals[i]);
        }
        return result;
    }

    /**
     * @see nenok.net.processor.Processor#retrieve(nenok.net.DefaultLocator)
     */
    
    public Valuation retrieve(DefaultLocator loc) throws RemoteException {
        Envelope env = table.get(loc.getEnvelopeID()); 
        return (env != null) ? env.getValuation() : null;
    }

    /**
     * @see nenok.net.processor.Processor#exists(net.jini.id.Uuid)
     */
    
    public boolean exists(Uuid envID) throws RemoteException {
        return table.containsKey(envID);
    }

    /**
     * @see nenok.net.processor.Processor#delete(nenok.net.DefaultLocator)
     */
    
    public void delete(DefaultLocator loc) throws RemoteException {
        table.remove(loc.getEnvelopeID()); 
    }

    /**
     * @see nenok.net.processor.Processor#reset()
     */
    
    public void reset() throws RemoteException {
        table.clear();
    }
    
    /**
     * @see nenok.net.processor.Processor#execute(nenok.net.task.Task)
     */
    
    public Locator execute(Task task) throws RemoteException, VAException {
        Envelope env = task.execute();
        store(env);
        return new DefaultLocator(pid, env);
    }

    /**
     * @see nenok.net.processor.Processor#getPID()
     */
    
    public Uuid getPID() throws RemoteException {
        return this.pid;
    }

    /**
     * @see nenok.net.processor.Processor#getCodebase()
     */
    
    public String getCodebase() throws RemoteException {
        return this.codebase;
    }

    /**
     * @see java.lang.Object#equals(java.lang.Object)
     */
    
    public boolean equals(Object o) {
        if(o instanceof Processor) {
            try {
                return ((Processor)o).getPID().equals(this.getPID());
            } catch (RemoteException e) {
                return false;
            }
        }
        return false;
    }
    
    /**
     * @see java.lang.Object#toString()
     */
    
    public String toString() {
        return "Processor "+pid;
    }
}
