package nenok.net.registry;

import java.rmi.Remote;
import java.rmi.RemoteException;
import java.util.Collection;

import nenok.net.Knowledgebase;
import nenok.net.Locator;

/**
 * The interface of the NENOK knowledgebase registry service. It extends the {@link Remote} interface 
 * such that its objects are available for remote calls. Note, that all methods within this class must 
 * throw a {@link RemoteException} by the RMI standard.
 * <br/><br/>
 * From this interface, the knowledgebase service proxy is generated automatically. It is a so-called 
 * dumb proxy - this means that no computation on client side is performed at all. Method calls are 
 * transmitted to the server.
 * 
 * @author Marc Pouly
 * @version 1.1
 */

public interface Registry extends Remote {

	/**
	 * Resets the knowledgebase registry. Deletes all content of the activ registry.
	 * @throws RemoteException Remote methods all throw remote exceptions.
	 */
	
	public void reset() throws RemoteException;
	
	/**
	 * Create a new knowledgebase with the given name.
	 * @param name The name of the new knowledgebase.
	 * @return <code>false</code> if a knowledgebase with this name exists already.
	 * @throws RemoteException Remote methods all throw remote exceptions.
	 */
	
	public boolean createKnowledgebase(String name) throws RemoteException;
	
	/**
	 * Deletes the knowledgebase with the given name.
	 * @param name The name of the knowledgebase to delete.
	 * @return <code>true</code> if a  knowledgebase has been deleted.
	 * @throws RemoteException Remote methods all throw remote exceptions.
	 */
	
	public boolean deleteKnowledgebase(String name) throws RemoteException;	
	
	/**
	 * Looks for a knowledgebase with the given name.
	 * @param name The name of the knowledgebase to find.
	 * @return The knowledgebase with the given name or <code>null</code>.
	 * @throws RemoteException Remote methods all throw remote exceptions.
	 */
	
	public Knowledgebase getKnowledgebase(String name) throws RemoteException;
	
    /**
     * @return The names of all registered knowledgebases.
     * @throws RemoteException Remote methods all throw remote exceptions.
     */
    
    public String[] getKnowledgebases() throws RemoteException;
    
	/**
	 * Adds a locator to the knowledgebase with the given name.
	 * @param knowledgebase The name of the knowledgebase.
	 * @param locator The locator to be added to the knowledgebase.
	 * @return <code>false</code> if a knowledgebase with the given name does
	 * not exist, <code>true</code>, if the element has been added successfully
     * @throws RemoteException Remote methods all throw remote exceptions.
	 */
	
	public boolean add(String knowledgebase, Locator locator) throws RemoteException;
	
	/**
	 * Adds a collection of locators to the knowledgebasewith the given name.
	 * @param knowledgebase The name of the knowledgebase.
	 * @param locators The collection of locators to add.
	 * @return <code>false</code> if a knowledgebase with the given name does
	 * not exist, <code>true</code>, if the knowledgebase changed as a result of the call.
	 * @throws RemoteException Remote methods all throw remote exceptions.
	 */
	
	public boolean addAll(String knowledgebase, Collection<Locator> locators) throws RemoteException;
    
    /**
     * Adds an array of locators to the knowledgebasewith the given name.
     * @param knowledgebase The name of the knowledgebase.
     * @param locators The array of locators to add.
     * @return <code>false</code> if a knowledgebase with the given name does
     * not exist, <code>true</code>, if the knowledgebase changed as a result of the call.
     * @throws RemoteException Remote methods all throw remote exceptions.
     */
    
    public boolean addAll(String knowledgebase, Locator[] locators) throws RemoteException;

	/**
	 * Removes a locator from the knowledgebase with the given name.
	 * @param knowledgebase The name of the knowledgebase.
	 * @param locator The locator to remove.
     * @return <code>true</code>, if the locator has been removed successfully.
	 * @throws RemoteException Remote methods all throw remote exceptions.
	 */
	
	public boolean remove(String knowledgebase, Locator locator) throws RemoteException;
		
	/**
	 * Does a knowledgebase with this name exist?
	 * @param knowledgebase The name of the knowledgebase.
	 * @return <code>true</code>, if a corresponding knowledgebase exist.
	 * @throws RemoteException Remote methods all throw remote exceptions.
	 */
	
	public boolean exists(String knowledgebase) throws RemoteException;
	
	/**
	 * Does the knowledgebase contain the given locator?
	 * @param knowledgebase The name of the knowledgebase.
	 * @param locator The locator to find.
	 * @return <code>true</code>, if the knowledgebase contains the given locator.
	 * @throws RemoteException Remote methods all throw remote exceptions.
	 */
	
	public boolean contains(String knowledgebase, Locator locator) throws RemoteException;
    
    /**
     * @return The IP of the host running this registry.
     * @throws RemoteException Remote methods all throw remote exceptions.
     */
    
    public String getIP() throws RemoteException;
}