package nenok.net.registry;

import java.io.IOException;
import java.net.UnknownHostException;
import java.rmi.RMISecurityManager;
import java.rmi.Remote;
import java.rmi.server.ExportException;
import java.util.logging.Level;
import java.util.logging.Logger;

import nenok.Services;
import net.jini.config.Configuration;
import net.jini.config.ConfigurationException;
import net.jini.config.ConfigurationProvider;
import net.jini.core.entry.Entry;
import net.jini.core.lookup.ServiceID;
import net.jini.discovery.DiscoveryGroupManagement;
import net.jini.discovery.LookupDiscoveryManager;
import net.jini.export.Exporter;
import net.jini.id.Uuid;
import net.jini.id.UuidFactory;
import net.jini.lease.LeaseRenewalManager;
import net.jini.lookup.JoinManager;
import net.jini.lookup.entry.Name;
import net.jini.lookup.entry.ServiceInfo;

import com.sun.jini.config.ConfigUtil;
import com.sun.jini.lookup.entry.BasicServiceType;
import com.sun.jini.start.LifeCycle;

/**
 * This class launches Nenok Registry instances and
 * registers them in a JINI lookup service.
 * 
 * @author Marc Pouly
 * @version 1.1
 */

public class RegistryStarter {
    
    private Exporter exporter;  
    private JoinManager joinMgr;
    
    /*
     * Implementation note: Do not delete this reference.
     * the garbage collector would delete the reference.
     */
    
    private RegistryImpl impl;
    private boolean loop = true;
        
    /*
     * Ensure that a SecurityManager is running.
     */
    
    private static Logger logger;
    
    static {
        if(System.getSecurityManager() == null) {
            System.setSecurityManager(new RMISecurityManager());
        }
        logger = Logger.getLogger(Registry.class.getPackage().toString());
        logger.setLevel(Level.INFO);
    }
        
    /**
     * Constructor:
     * @param configArgs Arguments from a configuration file.
     * @param lifeCycle no meaning - given by interface for unknown reasons.
     */
    
    public RegistryStarter(String[] configArgs, LifeCycle lifeCycle) {
        this(configArgs);
    }
    
    /**
     * Constructor:
     * @param args Array of configuration file names.
     */
    
    public RegistryStarter(String[] args) {
        
        /*
         * Service attributes:
         */
            
        Entry[] attributes = new Entry[] {new Name("Nenok Registry Service"), 
                                          new ServiceInfo("Nenok Registry Service", "Nenok Developer", "", "1.0", "", ""),
                                          new BasicServiceType("Nenok Registry")};
                
        Uuid uuid = UuidFactory.generate();
        ServiceID id = new ServiceID(uuid.getMostSignificantBits(), uuid.getLeastSignificantBits());
        
        /*
         * If a registry service is already running in this subset, abort !
         */
        
        if(Services.findRegistry() != null) {
        	logger.warning("A registry service is already running in this subnet. Start process aborted.");
        	System.exit(0);
        }
        
        /*
         * Get configuration from configuration file:
         */
        
        Configuration config = null;
        try {
            config = ConfigurationProvider.getInstance(args);
        } catch (ConfigurationException e) {
            logger.severe("Error while accessing configuration file.");
            e.printStackTrace();
            System.exit(1);
        }
        
        /*
         * Get exporter from configuration:
         */
        
        try {
            exporter = (Exporter)config.getEntry("Registry", "exporter", Exporter.class);
        } catch(ConfigurationException e) {
            logger.severe("Exception occured while accessing exporter entry from configuration file.");
            e.printStackTrace();
            System.out.println(2);
        }
            
        /*
         * Create proxy instance.
         */
        
        String host = "Unknown host";
        try {
            host = ConfigUtil.getHostName();
        } catch (UnknownHostException e1) {
            logger.warning("Unable to determine system's host name.");
        }
        
        impl = new RegistryImpl(host);
        
        Remote proxy = null;
        try {
            proxy = exporter.export(impl);
        } catch (ExportException e) {
            logger.severe("Unable to create proxy.");
            e.printStackTrace();
            System.exit(3);
        }
        
        if(proxy == null) {
            logger.severe("Unable to create proxy.");
            System.exit(4);            
        }
                
        /*
         * Register proxy:
         */
        
        try {
            LookupDiscoveryManager discover = new LookupDiscoveryManager(DiscoveryGroupManagement.ALL_GROUPS, null, null);
            joinMgr = new JoinManager(proxy, attributes, id, discover, new LeaseRenewalManager());
        } catch(IOException e) {
            e.printStackTrace();
            System.exit(3);
        }
        
        if(joinMgr == null) {
            logger.severe("Unable to create join manager.");
            System.exit(2); 
        }
                
        while(joinMgr.getJoinSet() == null || joinMgr.getJoinSet().length == 0) {
            try {
                Thread.sleep(100);
            } catch (InterruptedException e) {}
        }
        
        logger.info("Service successfully registered ...");
        
        /*
         * Loop:
         */
        
        while(loop) {
            try {
                Thread.sleep(1000);
            } catch (InterruptedException e) {}
        }
        
        logger.info("Service stopped ...");
     }
    
    /**
     * Terminates this service. This method unexports the registry service
     * and cancels all leases. The service disappears when the leases expire.
     * @return <code>true</code>, if the service has been unexported sucessfully. 
     */
    
    public boolean terminate() {
        this.loop = false;
        joinMgr.getLeaseRenewalManager().clear();
        return exporter.unexport(true);
    }
}