package nenok.va;

/**
 * This implementation of the valuation interface represents an identity element in a
 * valuation algebra. It has been shown that all valuation algebras can artificially be 
 * extended to have such an identity element.
 * <br/><br/>
 * Node, that identity elements are not weight predictable - they can be embedded in a 
 * weight predictable valuation algebra and inherit the weight predictor of the corresponding 
 * implementation.
 * 
 * @author Marc Pouly
 * @version 1.3
 */

public final class Identity implements Scalability, Idempotency {
	
	/**
	 * Unique instance of an identity element.
	 */
	
	public static final Identity INSTANCE = new Identity(); 
	
	/**
	 * Constructor:
	 */
	
	private Identity() {}
	
    /**
	 * @see nenok.va.Valuation#label()
	 */

	public Domain label() {
		return Domain.EMPTY;
	}
	
	/**
	 * @see nenok.va.Valuation#combine(nenok.va.Valuation)
	 */

	public Valuation combine(Valuation val) {
        return val;
	}
	
	/**
	 * The marginalization to the empty set is the only possible 
	 * marginalization for identity elements.
	 * @see nenok.va.Valuation#marginalize(nenok.va.Domain)
	 */

	public Valuation marginalize(Domain dom) throws VAException {
		if(dom.size() == 0) {
			return this;
		} 
		throw new VAException("Only marginalization to the empty domain is defined for identity elements.");	
	}
		
    /**
     * @see nenok.va.Valuation#weight()
     */
    
	public int weight() {
        return 0;
    }

	/**
	 * @return The string representation of identity elements.
	 */
	
	public String displayText() {
		return "Identity Element";
	}
	
    /**
	 * @see java.lang.Object#toString()
	 */
	
	public String toString() {
	    return displayText();
	}
	
	/**
	 * @see java.lang.Object#equals(java.lang.Object)
	 */
	
	public boolean equals(Object o) {
	    return (o instanceof Identity);
	}
	    
    /**
     * @see nenok.va.Scalability#scale()
     */
    
    public Scalability scale() {
        return this;
    }

    /**
     * @see nenok.va.Regularity#inverse()
     */
    
    public Regularity inverse() {
		return this;
    }
}
