package nenok.va;

import java.io.Serializable;

/**
 * This interface specifies a general valuation by its principal operations of 
 * labeling, marginalization &amp; combination. Ensure that all axioms of a 
 * valuation algebra are respected when implementing this interface. Valuation 
 * objects are serializable such that they can be transmitted between network hosts.
 * <br/><br/>
 * Additionally, this interface provides a method to copy valuation objects. This method
 * is mainly used by the implementation of {@link Identity} valuations. To give an example: 
 * the combination of an arbitrary valuation with an identity element should return a copy 
 * of the former. An alternative design strategy would have been to reuse the 
 * <code>Object.clone()</code> method. However, because this method is pre-implemented, the 
 * user is not forced to overwrite it. Additionally, the {@link Cloneable} interface in only a 
 * marker to ensure that <code>Object.clone()</code> can indeed be used. This is not satisfactory 
 * for our purposes.
 * <br/><br/>
 * The last method provides a generic weight function for valuations, which is used afterwards 
 * to model communication costs of local computation. Again, ensure that the properties of weight
 * functions are fulfilled.
 * 
 * @author Marc Pouly
 * @version 1.1
 */

public interface Valuation extends Serializable, Cloneable {
	
	/**
	 * Labeling of a valuation.
	 * @return The domain of the valuation
	 */
	
	public Domain label();
	
	/**
	 * Combination of valuations. Note that the method's argument is declared to be final 
     * to prevent modifications on this object.
	 * @param val The second valuation involved in the combination.
	 * @return The combination of <code>this</code> and <code>val</code>.
	 */
	
	public Valuation combine(Valuation val);
	
	/**
	 * Marginalization of a valuation onto a given domain.
	 * @param dom The domain onto this valuation is marginalized.
	 * @return The marginalization of <code>this</code> to <code>dom</code>.
	 * @throws VAException Exception occuring when a marginalization is performed onto an illegal domain.
	 */
	
	public Valuation marginalize(Domain dom) throws VAException;
        
    /**
     * @return The wieght of the current valuation.
     */
    
    public int weight();
}
