package nenok.va.semiring;

import java.io.Serializable;

import nenok.va.Domain;
import nenok.va.VAException;
import nenok.va.Variable;

/**
 * The semiring implementation uses configurations as fundamental computing element.
 * Configurations are arrays of elements of discrete variables. Therefore, this 
 * class represents a configuration by the following components:
 * <ul>
 *  <li>An array of discrete variables</li>
 *  <li>An array of values</li>
 * </ul>
 * Both arrays have the same size and the i-th value belongs to the i-th variable.
 * Configurations have three basic operations, namely: labeling, projection and combination.
 * These operations will induce their correspondances in a valuation algebra.
 * 
 * @author Marc Pouly
 * @version 1.0
 */

public class Configuration implements Serializable {
    
    private FiniteVariable[] vars;
    private Object[] values;
    
    /**
     * Constructor:
     * @param variables The variables of this configuration.
     * @param values The values corresponding to the array of variables. 
     * Value <code>i</code> belongs to variable <code>i</code>.
     */
    
    public Configuration(FiniteVariable[] variables, Object[] values) {
        if(variables.length != values.length) {
             throw new IllegalArgumentException("Argument arrays must have the same size.");
        }
        this.vars = variables;
        this.values = values;
    }
    
    /**
     * Implementation of the labeling procedure. Creates and returns a domain object 
     * containing the configuration's set of variables.
     * @return The configuration's domain.
     */
    
	@SuppressWarnings("cast") 
    public Domain label() {
        return new Domain((Variable[])vars);
    }
    
    /**
     * The operation of projection restricts a configuration to a given set of variables.
     * Each projection creates a new configuration instance. Note that the argument must 
     * be a subset of <code>this.label()</code>.
     * @param dom The domain onto the projection is performed.
     * @return The resulting configuration.
     * @throws VAException Exception thrown when projecting on illegal domain.
     */
    
    public Configuration project(Domain dom) throws VAException {
        if(!dom.subSetOf(this.label())) {
            throw new VAException("Illegal projection: "+dom+" is not a subset of "+this.label());
        }
        FiniteVariable[] newVars = new FiniteVariable[dom.size()];
        Object[] newValues = new Object[dom.size()];
        for (int i = 0, j = 0; i < values.length; i++) {
            if(dom.contains(vars[i])) {
                newVars[j] = vars[i];
                newValues[j++] = values[i];
            }
        }
        return new Configuration(newVars, newValues);   
    }
    
    /**
     * @see java.lang.Object#equals(java.lang.Object)
     */
    
    public boolean equals(Object c) {
        if(!(c instanceof Configuration)) {
            return false;
        }
        Configuration arg = (Configuration)c;       
        if(arg.vars.length != vars.length) {
            return false;
        }
        for (int i = 0; i < vars.length; i++) {
            int pos = arg.getIndex(vars[i]);
            if(pos < 0 || !values[i].equals(arg.values[pos])) {
                return false;
            }
        }
        return true;
    }
       
    /**
     * @return The size of this configuration.
     */
    
    public int size() {
        return values.length;
    }
    
    /**
     * This method computes the combination of two configurations. Each call of this method creates a
     * new configuration instance. The two initial configurations stay untouched.
     * @param c The second configuration to be combined with <code>this</code>.
     * @return The combined configuration as a new object.
     */
        
    public Configuration combine(Configuration c) {
        if(!this.isCombinable(c)) {
            throw new IllegalArgumentException("All intersecting values must be equal.");
        }
        
        // Create new instances:
        int size = vars.length + c.vars.length - countCommonVariables(c);
        FiniteVariable[] newVars = new FiniteVariable[size];
        Object[] newValues = new Object[size];
        
        // Fill new configuration arrays:
        int i = 0;
        for (; i < vars.length; i++) {
            newVars[i] = vars[i];
            newValues[i] = values[i];
        }
        for (int j = 0; j < c.vars.length; j++) {
            if(!this.contains(c.vars[j])) {
                newVars[i] = c.vars[j];
                newValues[i++] = c.values[j];
            }
        }
        return new Configuration(newVars, newValues);
    }
    
    /*
     * Helper methods:
     */
    
    /**
     * Is it possible to compute the combination of <code>this</code> and <code>c</code>.
     * @param c The second configuration.
     * @return <code>true</code> if the combination can be performed.
     */
    
    protected boolean isCombinable(Configuration c) {
        for (int i = 0; i < vars.length; i++) {
            for (int j = 0; j < c.vars.length; j++) {
                if(vars[i].equals(c.vars[j])) {
                    if(!(values[i].equals(c.values[j]))) {
                        return false;
                    }
                }
            }
        }
        return true;
    }
    
    /**
     * Tests if this configuration contains variable <code>var</code>.
     * @param var The variable to search.
     * @return <code>true</code> if this configuration contains the variable <code>var</code>.
     */
    
    private boolean contains(FiniteVariable var) {
        return getIndex(var) != -1;
    }
    
    /**
     * Get the index of <code>var</code> in the variable array.
     * @param var The variable to search.
     * @return The variable's index or <code>-1</code> if the variable cannot be found.
     */
    
    private int getIndex(FiniteVariable var) {
        for (int i = 0; i < vars.length; i++) {
            if(vars[i].equals(var)) {
                return i;
            }
        }
        return -1;
    }
    
    /**
     * Computes the number of (common) variables that are contained in both configurations.
     * @param c The second configuration.
     * @return The total number of variables contained in both configurations.
     */
    
    private int countCommonVariables(Configuration c) {
        int result = 0;
        for (int i = 0; i < vars.length; i++) {
            if(c.contains(vars[i])) {
                result++;
            }
        }
        return result;
    }
    
    /*
     * Output methods:
     */
    
    /**
     * @see java.lang.Object#toString()
     */
    
    public String toString() {
        return printVariables()+"\n"+printValues(); 
    }
    
    /**
     * Prints the variable array of this configuration.
     * @return The output string.
     */
    
    private String printVariables() {
        String result = "";
        for(int i = 0; i < vars.length; i++) {
            result += vars[i]+"\t";
        }
        return result;
    }
    
    /**
     * Prints the value array of this configuration.
     * @return The output string.
     */
    
    private String printValues() {
        String result = "";
        for (int i = 0; i < values.length; i++) {
            result += values[i]+"\t";
        }
        return result;
    }

    /*
     * Getter &amp; Setter:
     */
    
    /**
     * @return Returns the variable array of this configuration.
     */
    
    public FiniteVariable[] getVariables() {
        return vars;
    }

    /**
     * @return Returns the value array of this configuration. 
     */
    
    public Object[] getValues() {
        return values;
    }
}