package nenok.va.semiring;

import java.util.ArrayList;

import nenok.va.Domain;
import nenok.va.Identity;
import nenok.va.Predictability;
import nenok.va.Predictor;
import nenok.va.VAException;
import nenok.va.Valuation;

/**
 * Specifies a semiring induced valuation by leading its principal operations of 
 * labeling, marginalization &amp; combination back to the semiring operations.
 * Furthermore, all semiring induced valuation algebras possess a weight predictor. 
 * Therefore, this class extends {@link Predictability}.
 * 
 * @author Marc Pouly
 * @version 1.1
 */

public class SRValuation implements Predictability {
    
    protected Configuration[] configurations;
    protected Element[] elements;
           
    /*
     * Constructors:
     */
    
    /**
     * Constructor:
     */
    
    public SRValuation() {}
        
    /**
     * Constructor:
     * @param variables The variables of this valuation.
     * @param elements The semiring elements assigned to the configurations.
     */
   
    public SRValuation(FiniteVariable[] variables, Element[] elements) {
        // Compute the number of configurations.
        int size = 1;
        for (int i = 0; i < variables.length; i++) {
            size = size * variables[i].getFrame().length;
        }
        if(elements.length != size && !(variables.length == 0 && elements.length == 1)) {
            throw new IllegalArgumentException("Dimension of semiring element array doesn't match.");
        }
        this.elements = elements;
        
        // Create configurations:
        int width = variables.length;
        int height = size;
        int past = 1;
        String[][] matrix = new String[height][width];
        for (int i = 0; i < width; i++) {
            FiniteVariable var = variables[i];
            String[] frame = (String[])var.getFrame();
            int cardinality = frame.length;
            int repetitions = (height / cardinality) / past;
            for (int j = 0, pointer = 0; j < height; j++) {
                if(j > 0 && j % repetitions == 0) {
                    pointer++;
                    if(pointer >= frame.length) {
                        pointer = 0;
                    }
                }
                matrix[j][i] = frame[pointer];   
            }
            past *= cardinality;
        }
        configurations = new Configuration[height];
        for (int i = 0; i < height; i++) {
            String[] c = new String[width];
            for (int j = 0; j < width; j++) {
                c[j] = matrix[i][j];
            }
            configurations[i] = new Configuration(variables, c);
        }
    }
            
    /*
     * Valuation Algebra Operations:
     */
    
    /**
     * @see nenok.va.Valuation#label()
     */
    
    public Domain label() {
        return (configurations.length == 0) ? new Domain() : configurations[0].label();
    }
    
    /**
     * @see nenok.va.Valuation#combine(nenok.va.Valuation)
     */
    
    public Valuation combine(Valuation val) {
        if(val instanceof Identity) {
            return this;
        }
        if(!(val instanceof SRValuation)) {
            throw new IllegalArgumentException("Argument is not an instance of a semiring valuation.");
        }
        SRValuation arg = (SRValuation)val;
        
        // New instance fields:
        ArrayList<Configuration> newConfigs = new ArrayList<Configuration>();
        ArrayList<Element> newElts = new ArrayList<Element>();
        
        // Fill instance fields:
        for(int i = 0; i < configurations.length; i++) {
            Configuration c1 = this.configurations[i];
            Element v1 = this.elements[i];
            for(int j = 0; j < arg.configurations.length; j++) {
                Configuration c2 = arg.configurations[j];
                if(c1.isCombinable(c2)) {
                    Configuration c = c1.combine(c2);
                    Element v2 = arg.elements[j];
                    Element v = v1.multiply(v2);
                    newConfigs.add(newConfigs.size(), c);
                    newElts.add(newElts.size(), v);
                }
            }
        }
        
        // Build result valuation:
        SRValuation result = this.getInstance();
        result.configurations = newConfigs.toArray(new Configuration[newConfigs.size()]);
        result.elements = newElts.toArray(new Element[newElts.size()]);
        return result;
    }
    
    /**
     * @see nenok.va.Valuation#marginalize(nenok.va.Domain)
     */
    
    public Valuation marginalize(Domain dom) throws VAException {
        if(!dom.subSetOf(label())) {
            throw new VAException("Argument is not a subset of this valuation's domain.");
        }
        ArrayList<Configuration> newConfigs = new ArrayList<Configuration>();
        ArrayList<Element> newElts = new ArrayList<Element>();
        
        // Fill instance fields:
        for (int i = 0; i < this.configurations.length; i++) {
            Configuration c = configurations[i].project(dom);
            int index = newConfigs.indexOf(c);
            if(index >= 0) {
                newElts.set(index, newElts.get(index).add(elements[i])); 
            } else {
                newConfigs.add(newConfigs.size(), c);
                newElts.add(newElts.size(), elements[i]);
            }
        }
        
        // Build result valuation:
        SRValuation result = this.getInstance();
        result.configurations = newConfigs.toArray(new Configuration[newConfigs.size()]);
        result.elements = newElts.toArray(new Element[newElts.size()]);
        return result;
    }
                
    /**
     * @see nenok.va.Valuation#weight()
     */
    
    public int weight() {
        return predictor().predict(this.label());
    }
    
    /*
     * Predictablility:
     */
    
    /**
     * @see nenok.va.Predictability#predictor()
     */
    
    public Predictor predictor() {
        return SRPredictor.getInstance();
    }
    
    /*
     * Equality check:
     */

    /**
     * @see java.lang.Object#equals(java.lang.Object)
     */
    
    public boolean equals(Object o) {
        if(!(o instanceof SRValuation)) {
            return false;
        }
        SRValuation s = (SRValuation)o;
        if(!this.label().equals(s.label())) {
            return false;
        }           
        for (int i = 0; i < configurations.length; i++) {
            boolean found = false;
            for (int j = 0; j < s.configurations.length; j++) {
                if(configurations[i].equals(s.configurations[j])) {
                    found = true;
                    if(!elements[i].equals(s.elements[j])) {
                        return false;
                    }
                    break;
                }
            }
            if(!found) {
                return false;
            }
        }
        return true;
    }
    
    /*
     * Output methods:
     */
            
    /**
     * @see java.lang.Object#toString()
     */
    
    public String toString() {
        return displayConfigurations();
    }
    
    /**
     * @return The current semiring valuation as a string. 
     */
    
    public String displayConfigurations() {

            int max = 0;
            String format = "", result = "", line = "", title = "Value";
            
            if(configurations.length == 0) {
                return "";
            }
                        
            /*
             * Find string of maximum length:
             */
            
            for (int i = 0; i < configurations.length; i++) {
                
                // Variables:
                if(i == 0) {
                    for (int j = 0; j < configurations[i].size(); j++) {
                        FiniteVariable v = configurations[i].getVariables()[j];
                        if(max < v.toString().length()) {
                            max = v.toString().length();
                        }
                    }
                }
                
                // Semiring Values:
                for (int j = 0; j < configurations[i].size(); j++) {
                    Object v = configurations[i].getValues()[j];
                    if(max < v.toString().length()) {
                        max = v.toString().length();
                    }
                }                
            }
            
            if(max < title.length()) {
                max = title.length();
            }
            
            max += 2;
            
            /*
             * Prepare format string for title:
             */
            
            for (int i = 0; i < configurations[0].size()+1; i++) {
                format += "|%"+(i+1)+"$-"+max+"s";
            }
            format += "|\n";
            
            /*
             * Output header line:
             */
            
            Object[] data = new Object[configurations[0].size()+1];
            for (int i = 0; i < configurations[0].size(); i++) {
                data[i] = configurations[0].getVariables()[i];
            }
            data[configurations[0].getVariables().length] = title;
            result += String.format(format, data);
            
            /*
             * Horizontal line:
             */
            
            for (int i = 0; i < result.length()-1; i++) {
                line += "-";
            }
            line += "\n";
            result = line+result+line;

            /*
             * Prepare format string for configurations:
             */

            format = "";
            for (int i = 0; i < configurations[0].getVariables().length; i++) {
                format += "|%"+(i+1)+"$-"+max+"s";
            }
            format += "|%"+(configurations[0].getVariables().length+1)+"$-"+max+"s|\n";
            
            /*
             * Configurations:
             */
            
            for (int i = 0; i < configurations.length; i++) {
                Object row[] = new Object[configurations[i].getValues().length+1];
                for (int j = 0; j < configurations[i].getValues().length; j++) {
                    row[j] = configurations[i].getValues()[j];
                }
                row[row.length-1] = elements[i];
                result += String.format(format, row);
            }   
            
            result += line;
            
            return result;
    }

    /*
     * Helper methods:
     */
    
    /**
     * Invokes the empty constructor of the class of <code>this</code>.
     */
    
    private SRValuation getInstance() {
    	try {
			Class c = this.getClass();
			return (SRValuation)c.newInstance();
		} catch (InstantiationException e) {
			throw new UnsupportedOperationException("Internal Error: Semiring subclass cannot be instantiated.");
		} catch (IllegalAccessException e) {
			throw new UnsupportedOperationException("Internal Error: Semiring subclass cannot be instantiated.");
		}	
    }
}

